// DGen/SDL v1.15+
// Megadrive C++ module - misc memory

#include <stdio.h>
#ifdef WIN32
//#define _DEBUG_MEM
//#define _DEBUG_MEM2
//#define _DEBUG_MEM3
#include "../fmtest/stdafx.h"
LARGE_INTEGER  li1 , li2 ;
#endif
#include "md.h"
// md.h also has include fm.h

// This is to handle memory accesses outside of ROM and RAM
// e.g. c00000 and a00000 addresses
// REMEMBER NOT TO USE ANY STATIC variables, because they
// will exist thoughout ALL megadrives!

unsigned char md::z80_read(unsigned int a)
{
#ifdef _PSP10
	//*
	if     ( a <  0x2000 ){ return z80ram[a]   ; }
	else if( a == 0x4000 ){ return myfm_read(0); }
	else if( a == 0x4001 ){ return myfm_read(1); }
	else if( a == 0x4002 ){ return myfm_read(2); }
	else if( a == 0x4003 ){ return myfm_read(3); }
	else if( a >= 0x8000 )
	{ 
#ifdef _DEBUG_MEM3
		TRACE( "ZR:%08X\n", (z80_bank68k << 15) + (a & 0x7fff) );
#endif // _DEBUG_MEM3
		return misc_readbyte( (z80_bank68k << 15) + (a & 0x7fff) ); }
#ifdef _DEBUG_MEM2
	else { ASSERT(FALSE) ; }
#endif //_DEBUG_MEM2
	//*/

	/*
	switch( a >> 12 )
	{
	case 0 :
	case 1 : return z80ram[a];
	case 2 : break ;
	case 3 : break ;
	case 4 : return myfm_read( a & 3 );
	case 5 : break ;
	case 6 : break ;
	case 7 : break ;
	default: return misc_readbyte( (z80_bank68k << 15) + (a & 0x7fff) );
	}
	*/

#else //_PSP10
  if (a<0x2000) return z80ram[a&0xffff];


  if ((a&0xfffc)==0x4000) // 00 01 02 03 - not sure about reads
  {
    return myfm_read(a&3);
  }

  if (a>=0x8000)
  {
    int addr68k;
    addr68k=z80_bank68k<<15;
    addr68k+=a&0x7fff;

    return misc_readbyte(addr68k);
  }
#endif //_PSP10

  return 0;
}

unsigned short md::z80_port_read(unsigned short a)
{
  int ret=0;

//  dprintf("z80 port read %.2x return %.4x\n",a,ret);

  return ret;
}
void md::z80_port_write(unsigned short a,unsigned char v)
{
//  dprintf("z80 port write %.2x %.2x\n",a,v);
}


void md::z80_write(unsigned int a,unsigned char v)
{
#ifdef _PSP10

	//*
	if     ( a <  0x2000 ){ z80ram[a] = v    ; }
	else if( a == 0x4000 ){ myfm_write(0,v,1); }
	else if( a == 0x4001 ){ myfm_write(1,v,1); }
	else if( a == 0x4002 ){ myfm_write(2,v,1); }
	else if( a == 0x4003 ){ myfm_write(3,v,1); }
	else if( a == 0x6000 )
	{
		z80_bank68k>>=1;
		z80_bank68k+=(v&1)<<8;
		z80_bank68k&=0x1ff; // 9 bits and filled in the new top one
	}
	else if( a == 0x7f11 ){ mysn_write(v); }
	else if( a >= 0x8000 )
	{ 
#ifdef _DEBUG_MEM3
		TRACE( "ZW:%08X:%08X\n", (z80_bank68k << 15) + (a & 0x7fff), v );
#endif // _DEBUG_MEM3
		misc_writebyte( (z80_bank68k << 15) + (a & 0x7fff), v ); }
#ifdef _DEBUG_MEM2
	else { ASSERT(FALSE) ; }
#endif //_DEBUG_MEM2
	//*/

	/*
	switch( a >> 12 )
	{
	case 0 :
	case 1 : z80ram[a] = v ; break ;
	case 2 : break ;
	case 3 : break ;
	case 4 : myfm_write( a & 3, v, 1 ); break ;
	case 5 : break ;
	case 6 : 
		z80_bank68k>>=1;
		z80_bank68k+=(v&1)<<8;
		z80_bank68k&=0x1ff; // 9 bits and filled in the new top one
		break ;
	case 7 : mysn_write(v); break ;
	default: misc_writebyte( (z80_bank68k << 15) + (a & 0x7fff), v ); break ;
	}
	*/

#else //_PSP10
  if ((a&0xfffc)==0x4000) // 00 01 02 03
  {
    myfm_write(a&3,v,1);
    return ;
  }
  if (a==0x7f11)
  { mysn_write(v); return; }

  if (a==0x6000)
  {
    z80_bank68k>>=1;
    z80_bank68k+=(v&1)<<8;
    z80_bank68k&=0x1ff; // 9 bits and filled in the new top one
    return ;
  }
  if (a<0x2000) { z80ram[a&0xffff]=v; return; }


  if (a>=0x8000)
  {
    int addr68k;
    addr68k=z80_bank68k<<15;
    addr68k+=a&0x7fff;
    misc_writebyte(addr68k,v);
    return;
  }

//  dprintf("z80 write %.4x %.2x\n",a,v);
  return;
#endif //_PSP10

}

unsigned md::misc_readbyte(unsigned a)
{
#ifdef _DEBUG_MEM
	QueryPerformanceCounter( &li2 ) ;
	TRACE( "%10d:BR:%08X\n", li2.LowPart - li1.LowPart , a );
	::Sleep(1);
	li1 = li2 ;
#endif //_DEBUG_MEM

#ifdef _PSP10

//	if( a > 0x00ffffff ){ ASSERT(FALSE); }
//	a &= 0x00ffffff ;

	if ( a < MAX_ROM_SIZE ) //0xa00000) 
	{
		return rom[(a^1)];
	}
	//else if( a > 0x00e00000 ){ return ram[(a^1)&0xffff]; }
	else if( a > 0x00dfffff ){ return ram[(a^1)&0xffff]; }
	else if( a < 0x00a02000 )
	{ 
#ifdef _DEBUG_MEM3
	TRACE( "BR:%08X:%08X\n", a, z80ram[(a)&0x1fff] );
#endif //_DEBUG_MEM3

		return z80ram[(a)&0x1fff]; 
	}
	else
	{

	switch(a)
	{
	case 0x00c00000 :
	case 0x00c00001 : return vdp.readbyte() ;

	case 0x00a04000 : return myfm_read(0) ;
	case 0x00a04001 : return myfm_read(1) ;
	case 0x00a04002 : return myfm_read(2) ;
	case 0x00a04003 : return myfm_read(3) ;		

	case 0x00a10001 :

#if 0 //
		if ((country_ver&0xff0)==0xff0)
		{
		  // autodetect country
		  int headcode=0x80,avail=0;
		  int i;
		  for (i=0;i<3;i++)
		  {
			int ch=misc_readbyte(0x1f0+i);
				 if (ch=='U') avail|=1;
			else if (ch=='E') avail|=2;
			else if (ch=='J') avail|=4;
		  }
			   if (avail&1) headcode=0x80;
		  else if (avail&2) headcode=0x80;
		  else if (avail&4) headcode=0x00;
		  return headcode+(country_ver&0x0f) | (pal? 0x40 : 0);
		}
		return country_ver ;
#else //0
		return read_country_ver() ;
#endif // 0

	case 0x00a10002 :
	case 0x00a10003 :

		/*
		if (aoo3_six==3)
		{
		  // Extended pad info
		  if (aoo3_toggle==0) return ((pad[0]>>8)&0x30)+0x00;
		  else                return ((pad[0])&0x30)+0x40+((pad[0]>>16)&0xf);
		}
		else
		{
		  if (aoo3_toggle==0)
		  {
			if (aoo3_six==4) return ((pad[0]>>8)&0x30)+0x00+0x0f;
			else             return ((pad[0]>>8)&0x30)+0x00+(pad[0]&3);
		  }
		  else return ((pad[0])&0x30)+0x40+(pad[0]&15);
		}
		//*/

		//*
		return read_aoo3() ;
		//*/

		break ;

	case 0x00a10004 :
	case 0x00a10005 :
#if 0
		if (aoo5_six==3)
		{
		  // Extended pad info
		  if (aoo5_toggle==0x00) return ((pad[1]>>8)&0x30)+0x00;
		  else                   return ((pad[1])&0x30)+0x40+((pad[1]>>16)&0xf);
		}
		else
		{
		  if (aoo5_toggle==0x00)
		  {
			if (aoo5_six==4) return ((pad[1]>>8)&0x30)+0x00+0x0f;
			else             return ((pad[1]>>8)&0x30)+0x00+(pad[1]&3);
		  }
		  else return ((pad[1])&0x30)+0x40+(pad[1]&15);
		}
		break ; 
#else 
		return 0xFF ;
#endif //

	case 0x00a11100 : return z80_online ;

	case 0x00c00004 : coo4 ^= 0x03 ; return coo4 ;
	case 0x00c00005 : return coo5 | (pal? 1 : 0) ;

	case 0x00c00008 : return calculate_coo8() ;
	case 0x00c00009 : return calculate_coo9() ;

	case 0x00a11000 :
	case 0x00a11001 :
	case 0x00a11200 :
	case 0x00a11201 :
		return 0xFF ;

	case 0x00a130f1 :
		return save_active | save_prot ;

	case 0x00a1000F :
	case 0x00a10015 :
	case 0x00a1001B : return 0xFF ;

#ifdef _DEBUG_MEM2
	case 0x00a10006 :
	case 0x00a10007 :
	case 0x00a10008 :
	case 0x00a10009 :
	case 0x00a1000A :
	case 0x00a1000B :
	case 0x00A1000C :
	case 0x00A1000D :
	case 0x00a11101 :
		break ;
	default :
		TRACE( "BR:%08X\n", a );
		ASSERT(FALSE) ;
		break ;
#endif //_DEBUG_MEM
	}
	}

	return 0x00 ;

#else //_PSP10
  unsigned char ret=0;
  a&=0x00ffffff; // in case stars didn't clip to 24-bit bus


  // In case it's a rom read after all
  if (a<0xa00000) {
    // Saveram
    if((save_active)     && (save_len) &&
       (a >= save_start) && ((a - save_start) < save_len)) {
      return saveram[(a^1) - save_start];
    }
    else if ((signed)(a^1)<romlen) { ret=rom[(a^1)]; goto end; }
  }

  // In case it's a ram read after all
  if (a>=0xe00000) if (a<=0xffffff) { ret=ram[(a^1)&0xffff]; goto end; }

  // GFX data read (byte)
  if ((a&0xfffffffe)==0x00c00000) // 00 or 01
    { ret=vdp.readbyte(); goto end; }


  if ((a&0xfffffffc)==0xa04000)
  {
    ret=myfm_read(a&3); goto end;
  }

  if ((a>=0xa00000)&&(a<0xa08000))
  {
    // Read from z80's memory
    ret=z80_read(a&0x7fff); goto end;
  }

  // I/O port access
  if ((a&0xfffffe)==0xa10002)
  {
    if (aoo3_six==3)
    {
      // Extended pad info
      if (aoo3_toggle==0) ret=((pad[0]>>8)&0x30)+0x00;
      else ret=((pad[0])&0x30)+0x40+((pad[0]>>16)&0xf);
    }
    else
    {
      if (aoo3_toggle==0)
      {
        if (aoo3_six==4) ret=((pad[0]>>8)&0x30)+0x00+0x0f;
        else             ret=((pad[0]>>8)&0x30)+0x00+(pad[0]&3);
      }
      else ret=((pad[0])&0x30)+0x40+(pad[0]&15);
    }
    goto end;
  }

  if ((a&0xfffffe)==0xa10004)
  {
    if (aoo5_six==3)
    {
      // Extended pad info
      if (aoo5_toggle==0x00) ret=((pad[1]>>8)&0x30)+0x00;
      else ret=((pad[1])&0x30)+0x40+((pad[1]>>16)&0xf);
    }
    else
    {
      if (aoo5_toggle==0x00)
      {
        if (aoo5_six==4) ret=((pad[1]>>8)&0x30)+0x00+0x0f;
        else             ret=((pad[1]>>8)&0x30)+0x00+(pad[1]&3);
      }
      else ret=((pad[1])&0x30)+0x40+(pad[1]&15);
    }
    goto end;
  }

  if (a==0xa10001)
  {
    // overseas/pal/disk/0/ md version (0-f) - may make games act different!
    if ((country_ver&0xff0)==0xff0)
    {
      // autodetect country
      int headcode=0x80,avail=0;
      int i;
      for (i=0;i<3;i++)
      {
        int ch=misc_readbyte(0x1f0+i);
             if (ch=='U') avail|=1;
        else if (ch=='E') avail|=2;
        else if (ch=='J') avail|=4;
      }
           if (avail&1) headcode=0x80;
      else if (avail&2) headcode=0x80;
      else if (avail&4) headcode=0x00;
      ret=headcode+(country_ver&0x0f) | (pal? 0x40 : 0);
    }
    else ret=country_ver;
    goto end;
  }

  if (a==0xa11000) {ret=0xff; goto end; }
  if (a==0xa11001) {ret=0xff; goto end; }
  if (a==0xa11100)
  {
    ret=z80_online; goto end;
  }
  if (a==0xa11101) {ret=0x00; goto end; }
  if (a==0xa11200) {ret=0xff; goto end; }
  if (a==0xa11201) {ret=0xff; goto end; }

  // Genecyst style - toggles fifo full/empty
  if (a==0xc00004)
  {
    // This is the genecyst fudge for FIFO full/empty
    coo4^=0x03;
    ret=coo4; goto end;
  }
  if (a==0xc00005)
  {
    // This is the genecyst fudge for in h-blank
    //coo5^=0x04;
    ret=coo5 | (pal? 1 : 0); goto end;
  }

  if (a==0xc00008) { ret=calculate_coo8(); goto end; }
  if (a==0xc00009) { ret=calculate_coo9(); goto end; }


  end:
  return ret;
#endif //_PSP10
}

void md::misc_writebyte(unsigned a,unsigned d)
{
#ifdef _DEBUG_MEM
	QueryPerformanceCounter( &li2 ) ;
	TRACE( "%10d:BW:%08X:%08d\n", li2.LowPart - li1.LowPart, a, d );
	::Sleep(1);
	li1 = li2 ;
#endif //_DEBUG_MEM

#ifdef _PSP10

	if( a > 0x00dfffff ){ ram[(a^1)&0xffff] = d ; return ; }
	else if ( a < MAX_ROM_SIZE ) //0xa00000) 
	{
		save_update = 1 ;
		rom[(a^1)] = d ; 
		return ;
	}
	else if( a < 0x00a02000 )
	{ 
#ifdef _DEBUG_MEM3
	TRACE( "BW:%08X:%08X\n", a, d );
	::Sleep(1);
#endif //_DEBUG_MEM3
		z80ram[a&0x1fff] = d ;  return ; 
	}
	else 
	{

	switch(a)
	{
	case 0x00c00000 :
	case 0x00c00001 :
	case 0x00c00002 :
	case 0x00c00003 :
		vdp.writebyte(d);
		break ;

	case 0x00a07f11 :
	case 0x00c00011 :
	case 0x00c00017 :
		mysn_write(d);
		break ;

	case 0x00a04000 : myfm_write(0,d,1) ; break ;
	case 0x00a04001 : myfm_write(1,d,1) ; break ;
	case 0x00a04002 : myfm_write(2,d,1) ; break ;
	case 0x00a04003 : myfm_write(3,d,1) ; break ;
	case 0x00a06000 : write_z80_bank(d) ; break ;

	case 0x00a11100 :
		if (d==1)
		{
			run_to_odo_z80( odom - m68k_clks_left ) ;
			z80_online=0;
		}
		else
		{
			z80_online=1;
			odoz = odom - m68k_clks_left ;
		}
		break ;

	case 0x00a11200 :
		if (d==0)
		{ 
			//run_to_odo_z80( odo ) ;
			mz80reset(); YM2612ResetChip(); 
		}
		else
		{
			odoz = odom - m68k_clks_left ;
		}
		//else if( z80_reset ){ ; }
		z80_reset = d ;
		break ;
	case 0x00a11201 :
		break ;

	case 0x00a10003 :
		/*
		if (aoo3_six>=0 && (d&0x40)==0 && aoo3_toggle ) aoo3_six++;

		if (aoo3_six>0xc00000) aoo3_six&=~0x400000;
		// keep it circling around a high value

		if (d&0x40) aoo3_toggle=1; else aoo3_toggle=0;
		aoo3_six_timeout=0;
		*/


		//if( ((d & 0x40)==0) && ((aoo3_toggle & 0x40)==0) ){ aoo3_six = 1 ; }
		//if(d&0x20){ aoo3_six = 1 ; }

		aoo3_toggle = d ;
		break ;

	case 0x00a10005 :
		/*
		if (aoo5_six>=0 && (d&0x40)==0 && aoo5_toggle ) aoo5_six++;

		if (aoo5_six>0xc00000) aoo5_six&=~0x400000;
		// keep it circling around a high value

		if (d&0x40) aoo5_toggle=1; else aoo5_toggle=0;
		aoo5_six_timeout=0;
		*/
		break ;

	case 0x00a130f1 :
		write_save_flag( d ) ;
		break ;

#ifdef _DEBUG_MEM2
	case 0x00a10006 :
	case 0x00a10007 :
	case 0x00a10008 :
	case 0x00a10009 :
	case 0x00a1000A :
	case 0x00a1000B :
	case 0x00A1000C :
	case 0x00A1000D :
		break ;
	//case 0x00A11101: break ;
	default : TRACE( "BW:%08X:%08X\n", a, d ); /*ASSERT(FALSE);*/ break ;
#endif //_DEBUG_MEM
	}
	}

#else //_PSP10
  a&=0x00ffffff; // in case stars didn't clip to 24-bit bus

  // Saveram write
  if(save_len && !save_prot) if(a >= save_start)
    if ((a - save_start) < save_len)
      { saveram[(a^1) - save_start] = d; return; }

  // In case it's a ram write after all

  if (a>=0xe00000) if (a<=0xffffff) { ram[(a^1)&0xffff]=d; return; }
//  if (a>=0xe00000) if (a<=0xffffff) { ram[(a)&0xffff]=d; return; }

  // GFX data write (byte)
  if ((a&0xfffffffe)==0x00c00000) // 00 or 01
    { vdp.writebyte(d); return; }

  if ((a&0xfffffffc)==0xa04000)
  {
    myfm_write(a&3,d,1);
    return;
  }

  if (a==0xc00011)
  {
    mysn_write(d); return;
  }

  if (a==0xa11100)
  {
//d8 (W) 0: BUSREQ CANCEL
//       1: BUSREQ REQUEST
    if (d==1) z80_online=0;
    else z80_online=1;
    return;
  }
  if (a==0xa11101) return;
  if (a==0xa11200)
  {
    if (d==0)
      mz80reset();
    return;
  }
  if (a==0xa11201) return;

  if ((a>=0xa00000)&&(a<0xa08000))
  {
    // Write into z80's memory
    z80_write(a&0x7fff,d);
    return;
  }

  // I/O port access
  if (a==0xa10003)
  {
    if (aoo3_six>=0 && (d&0x40)==0 && aoo3_toggle ) aoo3_six++;

    if (aoo3_six>0xc00000) aoo3_six&=~0x400000;
    // keep it circling around a high value

    if (d&0x40) aoo3_toggle=1; else aoo3_toggle=0;
    aoo3_six_timeout=0;
    return;
  }
  if (a==0xa10005)
  {
    if (aoo5_six>=0 && (d&0x40)==0 && aoo5_toggle ) aoo5_six++;

    if (aoo5_six>0xc00000) aoo5_six&=~0x400000;
    // keep it circling around a high value

    if (d&0x40) aoo5_toggle=1; else aoo5_toggle=0;
    aoo5_six_timeout=0;
    return;
  }
  // Saveram status (thanks Steve :)
  if(a==0xa130f1)
  {
    // Bit 0: 0=rom active, 1=sram active
    // Bit 1: 0=writeable,  1=write protect
    save_active = d & 1;
    save_prot   = d & 2;
  }
#endif //_PSP10
}

unsigned md::misc_readword(unsigned a)
{
#ifdef _DEBUG_MEM
	if( (a % 0x1000) == 0 )
	{
	QueryPerformanceCounter( &li2 ) ;
	TRACE( "%10d:WR:%08X\n", li2.LowPart - li1.LowPart, a );
	::Sleep(1);
	li1 = li2 ;
	}
#endif //_DEBUG_MEM

#ifdef _PSP10
	if (     a < MAX_ROM_SIZE ){ return *((unsigned short*)&rom[(a)       ]); }
	else if( a > 0x00dfffff   ){ return *((unsigned short*)&ram[(a)&0xffff]); }
//	else if( a < 0x00a02000 ){ return *((unsigned short*)&z80ram[a&0x1fff]); }
	/*
	else if((a&0xfffffffc)==0x00c00000) // 00 or 02
	{
		return vdp.readword();
	}
	*/
	switch( a )
	{
	case 0x00c00000 :
	case 0x00c00002 : return vdp.readword() ;

	case 0x00c00004 :
		coo4 ^= 0x03 ;
		return (coo4 << 8) | (coo5 | (pal? 1 : 0)) ;

	case 0x00c00008 : return (calculate_coo8() << 8) | calculate_coo9() ;

	}

	return misc_readbyte(a)<<8 | misc_readbyte(a+1) ;

#else //_PSP10

  unsigned int ret=0;

  a&=0x00ffffff; // in case stars didn't clip to 24-bit bus

  if ((a&0xfffffffc)==0x00c00000) // 00 or 02
  {
    ret=vdp.readword();
    goto end;
  }

  // else pass onto readbyte
  ret =misc_readbyte(a)<<8;
  ret|=misc_readbyte(a+1);
  goto end;

  end:
  return ret;
#endif //_PSP10
}

void md::misc_writeword(unsigned a,unsigned d)
{
#ifdef _DEBUG_MEM
	QueryPerformanceCounter( &li2 ) ;
	TRACE( "%10d:WW:%08X:%08X\n", li2.LowPart - li1.LowPart, a, d );
	::Sleep(1);
	li1 = li2 ;
#endif //_DEBUG_MEM

#ifdef _PSP10
	//if( a > 0x00e00000 ){ *((unsigned short*)&ram[(a)&0xffff] ) = d ; return ; }
	if( a > 0x00dfffff ){ *((unsigned short*)&ram[(a)&0xffff] ) = d ; return ; }
	else if ( a < MAX_ROM_SIZE ) //0xa00000) 
	{
		save_update = 1 ;
		*((unsigned short*)&rom[a]) = d ; return ;
	}
//	else if( a < 0x00a02000 ){ *((unsigned short*)&z80ram[a&0x1fff]) = d ; return ; }
	else
	{

	switch( a )
	{
	case 0x00c00000 :
	case 0x00c00001 :
	case 0x00c00002 :
	case 0x00c00003 :
		vdp.writeword(d);
		break ;
	case 0x00c00004 :
	case 0x00c00005 :
	case 0x00c00006 :
	case 0x00c00007 :

		/*
		if (coo_waiting)
		{
			// Okay completed the vdp command
			coo_cmd|=d; coo_waiting=0; 

			vdp.command(coo_cmd);
			return;
		}
		if ((d&0xc000)==0x8000)
		{
			int addr; addr=d>>8; addr&=0x1f;
			if (vdp.reg[addr]!=(d&0xff))
			{
				// Store dirty information down to 1byte level in bits
				//int byt,bit;
				//byt=addr; bit=byt&7; byt>>=3; byt&=0x03;
				//vdp.dirt[0x30+byt]|=(1<<bit); 
				vdp.dirt[0x34]|=8;
			}
			vdp.reg[addr]=d&0xff;
			return;
		}
		coo_cmd=d<<16; coo_waiting=1;
		*/
		vdp_command( d ) ;
		return;
	case 0x00A11100 :
		if( d & 0x0100 )
		{
			run_to_odo_z80( odom - m68k_clks_left ) ;
			z80_online=0;
		}
		else
		{
			z80_online=1;
			odoz = odom - m68k_clks_left ;
		}
		break ;
	default :
		// else pass onto writebyte
		misc_writebyte(a   , d >> 8    );
		misc_writebyte(a+1 , d &  0xFF );
		break ;
	}
	}

#else //_PSP10
  a&=0x00ffffff; // in case stars didn't clip to 24-bit bus

  // GFX data write (word)
  if ((a&0xfffffffc)==0x00c00000) // 00 or 02
  {
    vdp.writeword(d);
    return;
  }

  if ((a&0xfffffffc)==0x00c00004) // 04 or 06
  {
    if (coo_waiting)
    {
      // Okay completed the vdp command
      coo_cmd|=d; coo_waiting=0; 

      vdp.command(coo_cmd);
      return;
    }
    if ((d&0xc000)==0x8000)
    {
      int addr; addr=d>>8; addr&=0x1f;
      if (vdp.reg[addr]!=(d&0xff))
      {
        // Store dirty information down to 1byte level in bits
        int byt,bit;
        byt=addr; bit=byt&7; byt>>=3; byt&=0x03;
        vdp.dirt[0x30+byt]|=(1<<bit); vdp.dirt[0x34]|=8;
      }
      vdp.reg[addr]=d&0xff;
      return;
    }
    coo_cmd=d<<16; coo_waiting=1;
    return;
  }

  // else pass onto writebyte
  misc_writebyte(a,d>>8);
  misc_writebyte(a+1,d&255);
#endif // _PSP10
}

unsigned md::misc_readlong(unsigned a)
{
#ifdef _DEBUG_MEM
	QueryPerformanceCounter( &li2 ) ;
	TRACE( "%10d:DR:%08X\n", li2.LowPart - li1.LowPart, a );
	::Sleep(1);
	li1 = li2 ;
#endif //_DEBUG_MEM

#ifdef _PSP10 
	unsigned long dd ;
	if( a % 4 == 0 )
	{
		if ( a < MAX_ROM_SIZE ) //0xa00000) 
		{
			dd = *((unsigned long*)&rom[(a)]) ;
			return (dd << 16) | (dd >> 16) ;
		}
		//else if( a > 0x00e00000 )
		else if( a > 0x00dfffff )
		{ 
			dd = *((unsigned long*)&ram[(a)&0xffff]) ;
			return (dd << 16) | (dd >> 16) ;	
		}
		/*
		else if( a < 0x00a02000 )
		{ 
			dd = *((unsigned long*)&z80ram[(a+0)&0x1fff]) ; 
			return (dd << 16) | (dd >> 16) ;	
		}
		*/
		else if((a&0xfffffffc)==0x00c00000) // 00 or 02
		{
			return vdp.readword()<<16 | vdp.readword() ;
		}
	}
	else
	{
		if ( a < MAX_ROM_SIZE ) //0xa00000) 
		{
			return *((unsigned short*)&rom[(a)]) << 16 | *((unsigned short*)&rom[(a+2)]);
		}
		//else if( a > 0x00e00000 )
		else if( a > 0x00dfffff )
		{ 
			return *((unsigned short*)&ram[(a)&0xffff]) << 16 | *((unsigned short*)&ram[(a+2)&0xffff] ); 
		}
		/*
		else if( a < 0x00a02000 )
		{ 
			return *((unsigned short*)&z80ram[(a+0)&0x1fff]) << 16 | *((unsigned short*)&z80ram[(a+2)&0x1fff]) ; 
		}
		//*/
		else if((a&0xfffffffc)==0x00c00000) // 00 or 02
		{
			return vdp.readword()<<16 | vdp.readword() ;
		}
	}
#endif 

	return misc_readword(a)<<16 | misc_readword(a+2) ;
}

void md::misc_writelong(unsigned a,unsigned d)
{
#ifdef _DEBUG_MEM
	QueryPerformanceCounter( &li2 ) ;
	TRACE( "%10d:DW:%08X:%08X\n", li2.LowPart - li1.LowPart, a, d );
	::Sleep(1);
	li1 = li2 ;
#endif //_DEBUG_MEM

#ifdef _PSP10
//	if( a > 0x00ffffff ){ ASSERT(FALSE); }

	if( a % 4 == 0 )
	{
		//if( a > 0x00e00000 )
		if( a > 0x00dfffff )
		{ 
			*((unsigned long*)&ram[(a)&0xffff]) = d >> 16 | d << 16 ;
			return ;
		}
		else if ( a < MAX_ROM_SIZE ) //0xa00000) 
		{
			save_update = 1 ;
			*((unsigned long*)&rom[(a)]) = d >> 16 | d << 16 ; 
			return ;
		}
	}
	else
	{
		//if( a > 0x00e00000 )
		if( a > 0x00dfffff )
		{ 
			*((unsigned short*)&ram[(a  )&0xffff]) = d >> 16           ;
			*((unsigned short*)&ram[(a+2)&0xffff]) = (unsigned short)d ;
			return ;
		}
		else if ( a < MAX_ROM_SIZE ) //0xa00000) 
		{
			save_update = 1 ;
			*((unsigned short*)&rom[(a  )]) = d >> 16           ;
			*((unsigned short*)&rom[(a+2)]) = (unsigned short)d ;
			return ;
		}
	}
#endif 

	switch( a )
	{
	case 0x00c00000 :
	case 0x00c00001 :
	case 0x00c00002 :
	case 0x00c00003 :
		vdp.writeword( d >> 16    );
		vdp.writeword( d & 0xFFFF );
		return ;
	case 0x00c00004 :
	case 0x00c00005 :
	case 0x00c00006 :
	case 0x00c00007 :
		vdp_command( d >> 16    ) ;
		vdp_command( d & 0xFFFF ) ;
		return;
	}

    misc_writeword(a  ,(d>>16)&0xffff);
    misc_writeword(a+2,      d&0xffff);
}


void md::vdp_command( unsigned d )
{
	if (coo_waiting)
	{
		// Okay completed the vdp command
		coo_cmd|=d; coo_waiting=0; 

		vdp.command(coo_cmd);
		return;
	}
	if ((d&0xC000)==0x8000)
	{
		vdp.dma_clear() ;
		int addr; addr=d>>8; addr&=0x1f;
		if (vdp.reg[addr]!=(d&0xff))
		{
			// Store dirty information down to 1byte level in bits
			//int byt,bit;
			//byt=addr; bit=byt&7; byt>>=3; byt&=0x03;
			//vdp.dirt[0x30+byt]|=(1<<bit); 
			vdp.dirt[0x34]|=8;
		}
		vdp.reg[addr]=d&0xff;
		return;
	}
	coo_cmd=d<<16; coo_waiting=1;
}

unsigned md::read_country_ver()
{
	// autodetect country
	int headcode=0x80,avail=0;
	int i;

	for (i=0;i<3;i++)
	{
		int ch=misc_readbyte(0x1f0+i);
		if      (ch=='U'){ avail |= 0x01; }
		else if (ch=='E'){ avail |= 0x02; }
		else if (ch=='J'){ avail |= 0x04; }
		else if( ch=='A'){ avail |= 0x10; }
		else if( ch=='B'){ avail |= 0x20; }
	}

	if      (avail & 0x10){ headcode=0x80; pal = 1 ; }
	else if (avail & 0x20){ headcode=0x00; pal = 1 ; }
	else 
	{
	  switch( country_ver )
	  {
	  default :
			   if (avail & 0x04){ headcode=0x00; pal = 0 ; }
		  else if (avail & 0x01){ headcode=0x80; pal = 0 ; }
		  else if (avail & 0x02){ headcode=0x80; pal = 1 ; }
		  else                  { headcode=0x00; pal = 0 ; }
		  break ;

	  case 1:
			   if (avail & 0x01){ headcode=0x80; pal = 0 ; }
		  else if (avail & 0x02){ headcode=0x80; pal = 1 ; }
		  else if (avail & 0x04){ headcode=0x00; pal = 0 ; }
		  else                  { headcode=0x80; pal = 0 ; }
		  break ;

	  case 2:
			   if (avail & 0x02){ headcode=0x80; pal = 1 ; }
		  else if (avail & 0x01){ headcode=0x80; pal = 0 ; }
		  else if (avail & 0x04){ headcode=0x00; pal = 0 ; }
		  else                  { headcode=0x80; pal = 1 ; }
		  break ;
	  }
	}

	return headcode | (pal? 0x40 : 0);
}

unsigned md::read_aoo3()
{
	if( aoo3_toggle & 0x40 )
	{
		if( aoo3_six )
		{
			aoo3_six = 0 ;
			return ((pad[0]   )&0x30)+0x40+((pad[0]>>16)&0xf);
		}
		else
		{
			if( aoo3_toggle & 0x20 ){ aoo3_six = 1 ; }

			return ((pad[0]   )&0x30)+0x40+( pad[0]     &0xf);
		}
	}
	else
	{
		if( aoo3_six )
		{
			return ((pad[0]>>8)&0x30)+0x00 ;
		}
		else
		{
			return ((pad[0]>>8)&0x30)+0x00+(pad[0]&3);
		}
	}

	return 0xFF ;
}

unsigned md::read_aoo5()
{
	if (aoo5_six==3)
	{
	  // Extended pad info
	  if (aoo5_toggle==0x00) return ((pad[1]>>8)&0x30)+0x00;
	  else                   return ((pad[1])&0x30)+0x40+((pad[1]>>16)&0xf);
	}
	else
	{
	  if (aoo5_toggle==0x00)
	  {
		if (aoo5_six==4) return ((pad[1]>>8)&0x30)+0x00+0x0f;
		else             return ((pad[1]>>8)&0x30)+0x00+(pad[1]&3);
	  }
	  else return ((pad[1])&0x30)+0x40+(pad[1]&15);
	}

	return 0 ;
}

void md::write_aoo3(unsigned d)
{
	/*
	if (aoo3_six>=0 && (d&0x40)==0 && aoo3_toggle ) aoo3_six++;

	if (aoo3_six>0xc00000) aoo3_six&=~0x400000;
	// keep it circling around a high value

	if (d&0x40) aoo3_toggle=1; else aoo3_toggle=0;
	aoo3_six_timeout=0;
	*/


	//if( ((d & 0x40)==0) && ((aoo3_toggle & 0x40)==0) ){ aoo3_six = 1 ; }
	//if(d&0x20){ aoo3_six = 1 ; }

	aoo3_toggle = d ;
}

void md::write_aoo5(unsigned d)
{
	if (aoo5_six>=0 && (d&0x40)==0 && aoo5_toggle ) aoo5_six++;

	if (aoo5_six>0xc00000) aoo5_six&=~0x400000;
	// keep it circling around a high value

	if (d&0x40) aoo5_toggle=1; else aoo5_toggle=0;
	aoo5_six_timeout=0;
}

void md::write_save_flag( unsigned d )
{
	int n ;

//	TRACE("%d, %d, %d, %d\n", d, save_active, save_prot, save_active );

	if( save_len > 0 )
	{
		if( (save_active && ((d & 1)==0)) || (!save_active && ((d & 1))) )
		{
			for( n=0; n < save_len ; n++ )
			{
				unsigned char buf   = rom[save_start + n] ;
				rom[save_start + n] = saveram[n]          ;
				saveram[n]          = buf                 ;
			}
		}

		save_active = d & 1;
		save_prot   = d & 2;
	}
}

void md::write_z80_bank( unsigned d )
{
	z80_bank68k>>=1;
	z80_bank68k+=(d&1)<<8;
	z80_bank68k&=0x1ff; // 9 bits and filled in the new top one
}
