// DGen/SDL v1.16+
// New raster effects engine
// I'd like to thank the Mac folks for giving me a good template to work from.
// This is just a cheap rehash of their code, except friendlier to other bit
// depths. :) I also put in a few little optimizations, like blank checking
// and especially the sprites.

#include <stdio.h>
#include <stdlib.h>
#include "md.h"

#undef WORDS_BIGENDIAN

// This is marked each time the palette is updated. Handy for the 8bpp
// implementation, so we don't waste time changing the palette unnecessarily.
int pal_dirty;

// Macros, to route draw_tile and draw_tile_solid to the right handler

#ifdef _PSP10

#define draw_tile(which, line, where)         draw_tile2((which),(line),(where))
#define draw_tile_solid(which, line, where)   draw_tile2_solid((which),(line),(where))
#else //_PSP10
#define draw_tile(which, line, where) \
	switch(Bpp)\
	  {\
	  case 1:\
	    draw_tile1((which),(line),(where)); break;\
	  case 2:\
	    draw_tile2((which),(line),(where)); break;\
	  case 3:\
	    draw_tile3((which),(line),(where)); break;\
	  case 4:\
	    draw_tile4((which),(line),(where)); break;\
	  }

#define draw_tile_solid(which, line, where) \
	switch(Bpp)\
	  {\
	  case 1:\
	    draw_tile1_solid((which),(line),(where)); break;\
	  case 2:\
	    draw_tile2_solid((which),(line),(where)); break;\
	  case 3:\
	    draw_tile3_solid((which),(line),(where)); break;\
	  case 4:\
	    draw_tile4_solid((which),(line),(where)); break;\
	  }
#endif //_PSP10

// Silly utility function, get a big-endian word
#ifdef WORDS_BIGENDIAN
static inline int get_word(unsigned char *where)
  { return (int)(*(unsigned short*)where); }
#else
#ifdef _PSP10

static inline int get_word(unsigned char *where) 
{
	unsigned short t = *((unsigned short*)where) ;
	t = (t << 8) | (t >> 8) ;
	return t ;
}

static inline unsigned long get_dword(unsigned char *where) 
{
	//*
	if( ((unsigned long)where) % 4 != 0 )
	{
		unsigned long t0 = *((unsigned short*)(where  )) ;
		unsigned long t1 = *((unsigned short*)(where+2)) ;
		t0 = t1 << 16 | t0 ;
		t0 = ((t0&0x00FF00FF) << 8) | ((t0&0xFF00FF00) >> 8) ;
		return t0 ;
	}
	//*/
	unsigned long t = *((unsigned long*)where) ;
	t = ((t&0x00FF00FF) << 8) | ((t&0xFF00FF00) >> 8) ;
	return t ;
}

#else //_PSP10
static inline int get_word(unsigned char *where) 
  { return (where[0] << 8) | where[1]; }
#endif //_PSP10
#endif

// Tile pixel masks
#ifdef WORDS_BIGENDIAN
#  define PIXEL0 (0xf0000000)
#  define PIXEL1 (0x0f000000)
#  define PIXEL2 (0x00f00000)
#  define PIXEL3 (0x000f0000)
#  define PIXEL4 (0x0000f000)
#  define PIXEL5 (0x00000f00)
#  define PIXEL6 (0x000000f0)
#  define PIXEL7 (0x0000000f)
#  define SHIFT0 (28)
#  define SHIFT1 (24)
#  define SHIFT2 (20)
#  define SHIFT3 (16)
#  define SHIFT4 (12)
#  define SHIFT5 ( 8)
#  define SHIFT6 ( 4)
#  define SHIFT7 ( 0)
#else // WORDS_BIGENDIAN
#  define PIXEL0 (0x000000f0)
#  define PIXEL1 (0x0000000f)
#  define PIXEL2 (0x0000f000)
#  define PIXEL3 (0x00000f00)
#  define PIXEL4 (0x00f00000)
#  define PIXEL5 (0x000f0000)
#  define PIXEL6 (0xf0000000)
#  define PIXEL7 (0x0f000000)
#  define SHIFT0 ( 4)
#  define SHIFT1 ( 0)
#  define SHIFT2 (12)
#  define SHIFT3 ( 8)
#  define SHIFT4 (20)
#  define SHIFT5 (16)
#  define SHIFT6 (28)
#  define SHIFT7 (24)
#endif // WORDS_BIGENDIAN

#ifdef ASM_TILES
extern "C" {
  void _asm_tiles_init(unsigned char *vram,
  		       unsigned char *reg,
		       unsigned *highpal);
  
  void _drawtile1(int which, int line, unsigned char *where);
  void _drawtile1_solid(int which, int line, unsigned char *where);
  void _drawtile2(int which, int line, unsigned char *where);
  void _drawtile2_solid(int which, int line, unsigned char *where);
  void _drawtile3(int which, int line, unsigned char *where);
  void _drawtile3_solid(int which, int line, unsigned char *where);
  void _drawtile4(int which, int line, unsigned char *where);
  void _drawtile4_solid(int which, int line, unsigned char *where);
}

// Pass off these calls to assembler counterparts
inline void md_vdp::draw_tile1_solid(int which, int line, unsigned char *where)
  { _drawtile1_solid(which, line, where); }

inline void md_vdp::draw_tile1(int which, int line, unsigned char *where)
  { _drawtile1(which, line, where); }

inline void md_vdp::draw_tile2_solid(int which, int line, unsigned char *where)
  { _drawtile2_solid(which, line, where); }

inline void md_vdp::draw_tile2(int which, int line, unsigned char *where)
  { _drawtile2(which, line, where); }

inline void md_vdp::draw_tile3_solid(int which, int line, unsigned char *where)
  { _drawtile3_solid(which, line, where); }

inline void md_vdp::draw_tile3(int which, int line, unsigned char *where)
  { _drawtile3(which, line, where); }

inline void md_vdp::draw_tile4_solid(int which, int line, unsigned char *where)
  { _drawtile4_solid(which, line, where); }

inline void md_vdp::draw_tile4(int which, int line, unsigned char *where)
  { _drawtile4(which, line, where); }

#else // ASM_TILES

// Blit tile solidly, for 1 byte-per-pixel
inline void md_vdp::draw_tile1_solid(int which, int line, unsigned char *where)
{
  unsigned tile, pal;

  pal = (which >> 9 & 0x30); // Determine which 16-color palette

  if(which & 0x1000) // y flipped
    line ^= 7; // take from the bottom, instead of the top

  if(reg[12] & 2) // interlace
    tile = *(unsigned*)(vram + ((which&0x7ff) << 6) + (line << 3));
  else
    tile = *(unsigned*)(vram + ((which&0x7ff) << 5) + (line << 2));

  // Blit the tile!
  if(which & 0x800) // x flipped
    {
      *(where  ) = ((tile & PIXEL7)>>SHIFT7) | pal;
      *(where+1) = ((tile & PIXEL6)>>SHIFT6) | pal;
      *(where+2) = ((tile & PIXEL5)>>SHIFT5) | pal;
      *(where+3) = ((tile & PIXEL4)>>SHIFT4) | pal;
      *(where+4) = ((tile & PIXEL3)>>SHIFT3) | pal;
      *(where+5) = ((tile & PIXEL2)>>SHIFT2) | pal;
      *(where+6) = ((tile & PIXEL1)>>SHIFT1) | pal;
      *(where+7) = ((tile & PIXEL0)>>SHIFT0) | pal;
    } else {
      *(where  ) = ((tile & PIXEL0)>>SHIFT0) | pal;
      *(where+1) = ((tile & PIXEL1)>>SHIFT1) | pal;
      *(where+2) = ((tile & PIXEL2)>>SHIFT2) | pal;
      *(where+3) = ((tile & PIXEL3)>>SHIFT3) | pal;
      *(where+4) = ((tile & PIXEL4)>>SHIFT4) | pal;
      *(where+5) = ((tile & PIXEL5)>>SHIFT5) | pal;
      *(where+6) = ((tile & PIXEL6)>>SHIFT6) | pal;
      *(where+7) = ((tile & PIXEL7)>>SHIFT7) | pal;
    }
} 

// Blit tile, leaving color zero transparent, for 1 byte per pixel
inline void md_vdp::draw_tile1(int which, int line, unsigned char *where)
{
  unsigned tile, pal;

  pal = (which >> 9 & 0x30); // Determine which 16-color palette

  if(which & 0x1000) // y flipped
    line ^= 7; // take from the bottom, instead of the top

  if(reg[12] & 2) // interlace
    tile = *(unsigned*)(vram + ((which&0x7ff) << 6) + (line << 3));
  else
    tile = *(unsigned*)(vram + ((which&0x7ff) << 5) + (line << 2));
  // If the tile is all 0's, why waste the time?
  if(!tile) return;

  // Blit the tile!
  if(which & 0x800) // x flipped
    {
      if(tile & PIXEL7) *(where  ) = ((tile & PIXEL7)>>SHIFT7) | pal;
      if(tile & PIXEL6) *(where+1) = ((tile & PIXEL6)>>SHIFT6) | pal;
      if(tile & PIXEL5) *(where+2) = ((tile & PIXEL5)>>SHIFT5) | pal;
      if(tile & PIXEL4) *(where+3) = ((tile & PIXEL4)>>SHIFT4) | pal;
      if(tile & PIXEL3) *(where+4) = ((tile & PIXEL3)>>SHIFT3) | pal;
      if(tile & PIXEL2) *(where+5) = ((tile & PIXEL2)>>SHIFT2) | pal;
      if(tile & PIXEL1) *(where+6) = ((tile & PIXEL1)>>SHIFT1) | pal;
      if(tile & PIXEL0) *(where+7) = ((tile & PIXEL0)>>SHIFT0) | pal;
    } else {
      if(tile & PIXEL0) *(where  ) = ((tile & PIXEL0)>>SHIFT0) | pal;
      if(tile & PIXEL1) *(where+1) = ((tile & PIXEL1)>>SHIFT1) | pal;
      if(tile & PIXEL2) *(where+2) = ((tile & PIXEL2)>>SHIFT2) | pal;
      if(tile & PIXEL3) *(where+3) = ((tile & PIXEL3)>>SHIFT3) | pal;
      if(tile & PIXEL4) *(where+4) = ((tile & PIXEL4)>>SHIFT4) | pal;
      if(tile & PIXEL5) *(where+5) = ((tile & PIXEL5)>>SHIFT5) | pal;
      if(tile & PIXEL6) *(where+6) = ((tile & PIXEL6)>>SHIFT6) | pal;
      if(tile & PIXEL7) *(where+7) = ((tile & PIXEL7)>>SHIFT7) | pal;
    }
} 

// Blit tile solidly, for 2 byte-per-pixel
inline void md_vdp::draw_tile2_solid(int which, int line, unsigned char *where)
{
  unsigned tile, temp ;
  unsigned short *pal;
  unsigned short *wwhere = (unsigned short*)where;

  pal = highpal + (which >> 9 & 0x30); // Determine which 16-color palette
  temp = *pal; *pal = highpal[reg[7]&0x3f]; // Get background color

  if(which & 0x1000) // y flipped
    line ^= 7; // take from the bottom, instead of the top

  if(reg[12] & 2) // interlace
    tile = *(unsigned*)(vram + ((which&0x7ff) << 6) + (line << 3));
  else
    tile = *(unsigned*)(vram + ((which&0x7ff) << 5) + (line << 2));

  // Blit the tile!
  if(which & 0x800) // x flipped
    {
      *(wwhere  ) = pal[((tile & PIXEL7)>>SHIFT7)];
      *(wwhere+1) = pal[((tile & PIXEL6)>>SHIFT6)];
      *(wwhere+2) = pal[((tile & PIXEL5)>>SHIFT5)];
      *(wwhere+3) = pal[((tile & PIXEL4)>>SHIFT4)];
      *(wwhere+4) = pal[((tile & PIXEL3)>>SHIFT3)];
      *(wwhere+5) = pal[((tile & PIXEL2)>>SHIFT2)];
      *(wwhere+6) = pal[((tile & PIXEL1)>>SHIFT1)];
      *(wwhere+7) = pal[((tile & PIXEL0)>>SHIFT0)];
    } else {
      *(wwhere  ) = pal[((tile & PIXEL0)>>SHIFT0)];
      *(wwhere+1) = pal[((tile & PIXEL1)>>SHIFT1)];
      *(wwhere+2) = pal[((tile & PIXEL2)>>SHIFT2)];
      *(wwhere+3) = pal[((tile & PIXEL3)>>SHIFT3)];
      *(wwhere+4) = pal[((tile & PIXEL4)>>SHIFT4)];
      *(wwhere+5) = pal[((tile & PIXEL5)>>SHIFT5)];
      *(wwhere+6) = pal[((tile & PIXEL6)>>SHIFT6)];
      *(wwhere+7) = pal[((tile & PIXEL7)>>SHIFT7)];
    }

  // Restore the original color
  *pal = temp;
} 

// Blit tile, leaving color zero transparent, for 2 byte per pixel
inline void md_vdp::draw_tile2(int which, int line, unsigned char *where)
{
  unsigned tile ;
  unsigned short *pal;
  unsigned short *wwhere = (unsigned short*)where;

  pal = highpal + (which >> 9 & 0x30); // Determine which 16-color palette

  if(which & 0x1000) // y flipped
    line ^= 7; // take from the bottom, instead of the top

  if(reg[12] & 2) // interlace
    tile = *(unsigned*)(vram + ((which&0x7ff) << 6) + (line << 3));
  else
    tile = *(unsigned*)(vram + ((which&0x7ff) << 5) + (line << 2));
  // If the tile is all 0's, why waste the time?
  if(!tile) return;

  // Blit the tile!
  if(which & 0x800) // x flipped
    {
      if(tile & PIXEL7) *(wwhere  ) = pal[((tile & PIXEL7)>>SHIFT7)];
      if(tile & PIXEL6) *(wwhere+1) = pal[((tile & PIXEL6)>>SHIFT6)];
      if(tile & PIXEL5) *(wwhere+2) = pal[((tile & PIXEL5)>>SHIFT5)];
      if(tile & PIXEL4) *(wwhere+3) = pal[((tile & PIXEL4)>>SHIFT4)];
      if(tile & PIXEL3) *(wwhere+4) = pal[((tile & PIXEL3)>>SHIFT3)];
      if(tile & PIXEL2) *(wwhere+5) = pal[((tile & PIXEL2)>>SHIFT2)];
      if(tile & PIXEL1) *(wwhere+6) = pal[((tile & PIXEL1)>>SHIFT1)];
      if(tile & PIXEL0) *(wwhere+7) = pal[((tile & PIXEL0)>>SHIFT0)];
    } else {
      if(tile & PIXEL0) *(wwhere  ) = pal[((tile & PIXEL0)>>SHIFT0)];
      if(tile & PIXEL1) *(wwhere+1) = pal[((tile & PIXEL1)>>SHIFT1)];
      if(tile & PIXEL2) *(wwhere+2) = pal[((tile & PIXEL2)>>SHIFT2)];
      if(tile & PIXEL3) *(wwhere+3) = pal[((tile & PIXEL3)>>SHIFT3)];
      if(tile & PIXEL4) *(wwhere+4) = pal[((tile & PIXEL4)>>SHIFT4)];
      if(tile & PIXEL5) *(wwhere+5) = pal[((tile & PIXEL5)>>SHIFT5)];
      if(tile & PIXEL6) *(wwhere+6) = pal[((tile & PIXEL6)>>SHIFT6)];
      if(tile & PIXEL7) *(wwhere+7) = pal[((tile & PIXEL7)>>SHIFT7)];
    }
} 

inline void md_vdp::draw_tile3_solid(int which, int line, unsigned char *where)
{
  unsigned tile; 
  unsigned char *pal;

  pal = ((unsigned char*)highpal) + (which >> 7 & 0xc0); // Determine 16-color palette

  if(which & 0x1000) // y flipped
    line ^= 7; // take from the bottom, instead of the top

  if(reg[12] & 2) // interlace
    tile = *(unsigned*)(vram + ((which&0x7ff) << 6) + (line << 3));
  else
    tile = *(unsigned*)(vram + ((which&0x7ff) << 5) + (line << 2));

  // Blit the tile (this is UGLY! Do you know of a better way? :P)
  if(which & 0x800)
    {
      *(where   ) = pal[((tile & 0x0f000000)>>22)+1];
      *(where+1 ) = pal[((tile & 0x0f000000)>>22)+2];
      *(where+2 ) = pal[((tile & 0x0f000000)>>22)  ];
      *(where+3 ) = pal[((tile & 0xf0000000)>>26)+1];
      *(where+4 ) = pal[((tile & 0xf0000000)>>26)+2];
      *(where+5 ) = pal[((tile & 0xf0000000)>>26)  ];
      *(where+6 ) = pal[((tile & 0x000f0000)>>14)+1];
      *(where+7 ) = pal[((tile & 0x000f0000)>>14)+2];
      *(where+8 ) = pal[((tile & 0x000f0000)>>14)  ];
      *(where+9 ) = pal[((tile & 0x00f00000)>>18)+1];
      *(where+10) = pal[((tile & 0x00f00000)>>18)+2];
      *(where+11) = pal[((tile & 0x00f00000)>>18)  ];
      *(where+12) = pal[((tile & 0x00000f00)>> 6)+1];
      *(where+13) = pal[((tile & 0x00000f00)>> 6)+2];
      *(where+14) = pal[((tile & 0x00000f00)>> 6)  ];
      *(where+15) = pal[((tile & 0x0000f000)>>10)+1];
      *(where+16) = pal[((tile & 0x0000f000)>>10)+2];
      *(where+17) = pal[((tile & 0x0000f000)>>10)  ];
      *(where+18) = pal[((tile & 0x0000000f)<< 2)+1];
      *(where+19) = pal[((tile & 0x0000000f)<< 2)+2];
      *(where+20) = pal[((tile & 0x0000000f)<< 2)  ];
      *(where+21) = pal[((tile & 0x000000f0)>> 2)+1];
      *(where+22) = pal[((tile & 0x000000f0)>> 2)+2];
      *(where+23) = pal[((tile & 0x000000f0)>> 2)  ];
    } else {
      *(where   ) = pal[((tile & 0x000000f0)>> 2)+1];
      *(where+ 1) = pal[((tile & 0x000000f0)>> 2)+2];
      *(where+ 2) = pal[((tile & 0x000000f0)>> 2)  ];
      *(where+ 3) = pal[((tile & 0x0000000f)<< 2)+1];
      *(where+ 4) = pal[((tile & 0x0000000f)<< 2)+2];
      *(where+ 5) = pal[((tile & 0x0000000f)<< 2)  ];
      *(where+ 6) = pal[((tile & 0x0000f000)>>10)+1];
      *(where+ 7) = pal[((tile & 0x0000f000)>>10)+2];
      *(where+ 8) = pal[((tile & 0x0000f000)>>10)  ];
      *(where+ 9) = pal[((tile & 0x00000f00)>> 6)+1];
      *(where+10) = pal[((tile & 0x00000f00)>> 6)+2];
      *(where+11) = pal[((tile & 0x00000f00)>> 6)  ];
      *(where+12) = pal[((tile & 0x00f00000)>>18)+1];
      *(where+13) = pal[((tile & 0x00f00000)>>18)+2];
      *(where+14) = pal[((tile & 0x00f00000)>>18)  ];
      *(where+15) = pal[((tile & 0x000f0000)>>14)+1];
      *(where+16) = pal[((tile & 0x000f0000)>>14)+2];
      *(where+17) = pal[((tile & 0x000f0000)>>14)  ];
      *(where+18) = pal[((tile & 0xf0000000)>>26)+1];
      *(where+19) = pal[((tile & 0xf0000000)>>26)+2];
      *(where+20) = pal[((tile & 0xf0000000)>>26)  ];
      *(where+21) = pal[((tile & 0x0f000000)>>22)+1];
      *(where+22) = pal[((tile & 0x0f000000)>>22)+2];
      *(where+23) = pal[((tile & 0x0f000000)>>22)  ];
    }
}

inline void md_vdp::draw_tile3(int which, int line, unsigned char *where)
{
  unsigned tile; 
  unsigned char *pal;

  pal = ((unsigned char*)highpal) + (which >> 7 & 0xc0); // Determine 16-color palette

  if(which & 0x1000) // y flipped
    line ^= 7; // take from the bottom, instead of the top

  if(reg[12] & 2) // interlace
    tile = *(unsigned*)(vram + ((which&0x7ff) << 6) + (line << 3));
  else
    tile = *(unsigned*)(vram + ((which&0x7ff) << 5) + (line << 2));
  // If it's empty, why waste the time?
  if(!tile) return;

  // Blit the tile (this is UGLY! Do you know of a better way? :P)
  if(which & 0x800)
    {
      // The numbers are a little wierd, because of little endian. *gag*
      // <SARCASM>Thanks Intel!</SARCASM>
      if(tile & 0x0f000000) { *(where   ) = pal[((tile & 0x0f000000)>>22)+1];
      			      *(where+1 ) = pal[((tile & 0x0f000000)>>22)+2];
      			      *(where+2 ) = pal[((tile & 0x0f000000)>>22)  ]; }
      if(tile & 0xf0000000) { *(where+3 ) = pal[((tile & 0xf0000000)>>26)+1];
      			      *(where+4 ) = pal[((tile & 0xf0000000)>>26)+2];
      			      *(where+5 ) = pal[((tile & 0xf0000000)>>26)  ]; }
      if(tile & 0x000f0000) { *(where+6 ) = pal[((tile & 0x000f0000)>>14)+1];
      			      *(where+7 ) = pal[((tile & 0x000f0000)>>14)+2];
      			      *(where+8 ) = pal[((tile & 0x000f0000)>>14)  ]; }
      if(tile & 0x00f00000) { *(where+9 ) = pal[((tile & 0x00f00000)>>18)+1];
      			      *(where+10) = pal[((tile & 0x00f00000)>>18)+2];
      			      *(where+11) = pal[((tile & 0x00f00000)>>18)  ]; }
      if(tile & 0x00000f00) { *(where+12) = pal[((tile & 0x00000f00)>> 6)+1];
      			      *(where+13) = pal[((tile & 0x00000f00)>> 6)+2];
      			      *(where+14) = pal[((tile & 0x00000f00)>> 6)  ]; }
      if(tile & 0x0000f000) { *(where+15) = pal[((tile & 0x0000f000)>>10)+1];
      			      *(where+16) = pal[((tile & 0x0000f000)>>10)+2];
      			      *(where+17) = pal[((tile & 0x0000f000)>>10)  ]; }
      if(tile & 0x0000000f) { *(where+18) = pal[((tile & 0x0000000f)<< 2)+1];
      			      *(where+19) = pal[((tile & 0x0000000f)<< 2)+2];
      			      *(where+20) = pal[((tile & 0x0000000f)<< 2)  ]; }
      if(tile & 0x000000f0) { *(where+21) = pal[((tile & 0x000000f0)>> 2)+1];
      			      *(where+22) = pal[((tile & 0x000000f0)>> 2)+2];
      			      *(where+23) = pal[((tile & 0x000000f0)>> 2)  ]; }
    } else {
      if(tile & 0x000000f0) { *(where   ) = pal[((tile & 0x000000f0)>> 2)+1];
      			      *(where+ 1) = pal[((tile & 0x000000f0)>> 2)+2];
      			      *(where+ 2) = pal[((tile & 0x000000f0)>> 2)  ]; }
      if(tile & 0x0000000f) { *(where+ 3) = pal[((tile & 0x0000000f)<< 2)+1];
      			      *(where+ 4) = pal[((tile & 0x0000000f)<< 2)+2];
      			      *(where+ 5) = pal[((tile & 0x0000000f)<< 2)  ]; }
      if(tile & 0x0000f000) { *(where+ 6) = pal[((tile & 0x0000f000)>>10)+1];
      			      *(where+ 7) = pal[((tile & 0x0000f000)>>10)+2];
      			      *(where+ 8) = pal[((tile & 0x0000f000)>>10)  ]; }
      if(tile & 0x00000f00) { *(where+ 9) = pal[((tile & 0x00000f00)>> 6)+1];
      			      *(where+10) = pal[((tile & 0x00000f00)>> 6)+2];
      			      *(where+11) = pal[((tile & 0x00000f00)>> 6)  ]; }
      if(tile & 0x00f00000) { *(where+12) = pal[((tile & 0x00f00000)>>18)+1];
      			      *(where+13) = pal[((tile & 0x00f00000)>>18)+2];
      			      *(where+14) = pal[((tile & 0x00f00000)>>18)  ]; }
      if(tile & 0x000f0000) { *(where+15) = pal[((tile & 0x000f0000)>>14)+1];
      			      *(where+16) = pal[((tile & 0x000f0000)>>14)+2];
      			      *(where+17) = pal[((tile & 0x000f0000)>>14)  ]; }
      if(tile & 0xf0000000) { *(where+18) = pal[((tile & 0xf0000000)>>26)+1];
      			      *(where+19) = pal[((tile & 0xf0000000)>>26)+2];
      			      *(where+20) = pal[((tile & 0xf0000000)>>26)  ]; }
      if(tile & 0x0f000000) { *(where+21) = pal[((tile & 0x0f000000)>>22)+1];
      			      *(where+22) = pal[((tile & 0x0f000000)>>22)+2];
      			      *(where+23) = pal[((tile & 0x0f000000)>>22)  ]; }
    }
}

// Blit tile solidly, for 4 byte-per-pixel
inline void md_vdp::draw_tile4_solid(int which, int line, unsigned char *where)
{
  unsigned tile;
  unsigned short temp, *pal;
  unsigned *wwhere = (unsigned*)where;

  pal = highpal + (which >> 9 & 0x30); // Determine which 16-color palette
  temp = *pal; *pal = highpal[reg[7]&0x3f]; // Get background color

  if(which & 0x1000) // y flipped
    line ^= 7; // take from the bottom, instead of the top

  if(reg[12] & 2) // interlace
    tile = *(unsigned*)(vram + ((which&0x7ff) << 6) + (line << 3));
  else
    tile = *(unsigned*)(vram + ((which&0x7ff) << 5) + (line << 2));

  // Blit the tile!
  if(which & 0x800) // x flipped
    {
      *(wwhere  ) = pal[((tile & PIXEL7)>>SHIFT7)];
      *(wwhere+1) = pal[((tile & PIXEL6)>>SHIFT6)];
      *(wwhere+2) = pal[((tile & PIXEL5)>>SHIFT5)];
      *(wwhere+3) = pal[((tile & PIXEL4)>>SHIFT4)];
      *(wwhere+4) = pal[((tile & PIXEL3)>>SHIFT3)];
      *(wwhere+5) = pal[((tile & PIXEL2)>>SHIFT2)];
      *(wwhere+6) = pal[((tile & PIXEL1)>>SHIFT1)];
      *(wwhere+7) = pal[((tile & PIXEL0)>>SHIFT0)];
    } else {
      *(wwhere  ) = pal[((tile & PIXEL0)>>SHIFT0)];
      *(wwhere+1) = pal[((tile & PIXEL1)>>SHIFT1)];
      *(wwhere+2) = pal[((tile & PIXEL2)>>SHIFT2)];
      *(wwhere+3) = pal[((tile & PIXEL3)>>SHIFT3)];
      *(wwhere+4) = pal[((tile & PIXEL4)>>SHIFT4)];
      *(wwhere+5) = pal[((tile & PIXEL5)>>SHIFT5)];
      *(wwhere+6) = pal[((tile & PIXEL6)>>SHIFT6)];
      *(wwhere+7) = pal[((tile & PIXEL7)>>SHIFT7)];
    }
  // Restore the original color
  *pal = temp;
} 

// Blit tile, leaving color zero transparent, for 4 byte per pixel
inline void md_vdp::draw_tile4(int which, int line, unsigned char *where)
{
  unsigned tile;
  unsigned short *pal;
  unsigned *wwhere = (unsigned*)where;

  pal = highpal + (which >> 9 & 0x30); // Determine which 16-color palette

  if(which & 0x1000) // y flipped
    line ^= 7; // take from the bottom, instead of the top

  if(reg[12] & 2) // interlace
    tile = *(unsigned*)(vram + ((which&0x7ff) << 6) + (line << 3));
  else
    tile = *(unsigned*)(vram + ((which&0x7ff) << 5) + (line << 2));
  // If the tile is all 0's, why waste the time?
  if(!tile) return;

  // Blit the tile!
  if(which & 0x800) // x flipped
    {
      if(tile & PIXEL7) *(wwhere  ) = pal[((tile & PIXEL7)>>SHIFT7)];
      if(tile & PIXEL6) *(wwhere+1) = pal[((tile & PIXEL6)>>SHIFT6)];
      if(tile & PIXEL5) *(wwhere+2) = pal[((tile & PIXEL5)>>SHIFT5)];
      if(tile & PIXEL4) *(wwhere+3) = pal[((tile & PIXEL4)>>SHIFT4)];
      if(tile & PIXEL3) *(wwhere+4) = pal[((tile & PIXEL3)>>SHIFT3)];
      if(tile & PIXEL2) *(wwhere+5) = pal[((tile & PIXEL2)>>SHIFT2)];
      if(tile & PIXEL1) *(wwhere+6) = pal[((tile & PIXEL1)>>SHIFT1)];
      if(tile & PIXEL0) *(wwhere+7) = pal[((tile & PIXEL0)>>SHIFT0)];
    } else {
      if(tile & PIXEL0) *(wwhere  ) = pal[((tile & PIXEL0)>>SHIFT0)];
      if(tile & PIXEL1) *(wwhere+1) = pal[((tile & PIXEL1)>>SHIFT1)];
      if(tile & PIXEL2) *(wwhere+2) = pal[((tile & PIXEL2)>>SHIFT2)];
      if(tile & PIXEL3) *(wwhere+3) = pal[((tile & PIXEL3)>>SHIFT3)];
      if(tile & PIXEL4) *(wwhere+4) = pal[((tile & PIXEL4)>>SHIFT4)];
      if(tile & PIXEL5) *(wwhere+5) = pal[((tile & PIXEL5)>>SHIFT5)];
      if(tile & PIXEL6) *(wwhere+6) = pal[((tile & PIXEL6)>>SHIFT6)];
      if(tile & PIXEL7) *(wwhere+7) = pal[((tile & PIXEL7)>>SHIFT7)];
    }
} 
#endif // ASM_TILES

// Draw the window (front or back)
void md_vdp::draw_window(int line, int front)
{
	int size;
	int x, y, w, start;
	int pl, add;
	int total_window;
	unsigned char *where;
	int which;
	// Set everything up
	y = line >> 3;
	total_window = (y < (reg[18]&0x1f)) ^ (reg[18] >> 7);

	// Wide or narrow
	size = (reg[12] & 1)? 64 : 32;

	pl = (reg[3] << 10) + ((y&0x3f)*size*2);

	// Wide(320) or narrow(256)?
	if(reg[12] & 1)
	{
		w = 40;
		start = -8;
	} else {
		w = 32;
		start = 24;
	}
	add = -2;
	where = dest + (start * Bpp);
	for(x=-1; x<w; ++x)
	{
		if(!total_window)
			if(reg[17] & 0x80) 
			{
				if(x < ((reg[17]&0x1f) << 1)) goto skip;
			} else {
				if(x >= ((reg[17]&0x1f) << 1)) goto skip;
			}
		which = get_word(((unsigned char*)vram) + (pl+(add&((size-1)<<1))));

		if((which>>15) == front)
			draw_tile(which, line&7, where);
skip:
		add += 2; where += Bpp_times8;
	}
}

void md_vdp::draw_sprites(int line, int front)
{
	unsigned which;
	int tx, ty, x, y, xend, ysize, yoff, i;
	unsigned char *where, *sprite;
	// Sprites have to be in reverse order :P
	for(i = sprite_count - 1; i >= 0; --i)
	{
		sprite = sprite_base + (sprite_order[i] << 3);
		// Get the first tile
		which = get_word(sprite + 4);
		// Only do it if it's on the right priority
		if((which >> 15) == front)
		{
			// Get the sprite's location
			y = get_word(sprite);
			x = get_word(sprite + 6) & 0x1ff;

			// Interlace?
			if(reg[12] & 2)
				y = (y & 0x3fe) >> 1;
			else
				y &= 0x1ff;

			x -= 0x80;
			y -= 0x80;
			yoff = (line - y);

			// Narrow mode?
			if(!(reg[12] & 1)) x += 32;

			xend = ((sprite[2] << 1) & 0x18) + x;
			ysize = sprite[2] & 0x3;

			// Render if this sprite's on this line
			if(xend > -8 && x < 320 && yoff >= 0 && yoff <= (ysize<<3)+7)
			{
				ty = yoff & 7;
				// y flipped?
				if(which & 0x1000)
					which += ysize - (yoff >> 3);
				else
					which += (yoff >> 3);
				++ysize;
				// x flipped?
				if(which & 0x800)
				{
					where = dest + (xend * Bpp);
					for(tx = xend; tx >= x; tx -= 8)
					{
						if(tx > -8 && tx < 320)
							draw_tile(which, ty, where);
						which += ysize;
						where -= Bpp_times8;
					}
				} 
				else 
				{
					where = dest + (x * Bpp);
					for(tx = x; tx <= xend; tx += 8)
					{
						if(tx > -8 && tx < 320)
							draw_tile(which, ty, where);
						which += ysize;
						where += Bpp_times8;
					}
				} 
			}
		}
	}
}

// The body for the next few functions is in an extraneous header file.
// Phil, I hope I left enough in this file for GLOBAL to hack it right. ;)
// Thanks to John Stiles for this trick :)

void md_vdp::draw_plane_back0(int line)
{
#define FRONT 0
#define PLANE 0
#include "ras-drawplane.h"
#undef PLANE
#undef FRONT
}

void md_vdp::draw_plane_back1(int line)
{
#define FRONT 0
#define PLANE 1
#include "ras-drawplane.h"
#undef PLANE
#undef FRONT
}

void md_vdp::draw_plane_front0(int line)
{
#define FRONT 1
#define PLANE 0
#include "ras-drawplane.h"
#undef PLANE
#undef FRONT
}

void md_vdp::draw_plane_front1(int line)
{
#define FRONT 1
#define PLANE 1
#include "ras-drawplane.h"
#undef PLANE
#undef FRONT
}



////////////////////////////////////////////////////////////////////////////////////////////
/// for PSP
////////////////////////////////////////////////////////////////////////////////////////////
#ifdef _PSP10

//#define _USE_LAY_TBL
#ifdef  _USE_LAY_TBL
#include "_lay_tbl.cpp"

inline void md_vdp::draw_lay_fast( unsigned long which, unsigned long tile, unsigned char* palnum )
{
	if( tile )
	{
		if( which & 0x0800 ) // x flipped
		{
			if( (int)palnum % 2 )
			{
			//*
				which = ((which >> 9) & 0x0030) ;
				if( tile & PIXEL7 ){ *( palnum    ) = ((tile & PIXEL7) >> SHIFT7) + which ; }
				if( tile & PIXEL6 ){ *( palnum +1 ) = ((tile & PIXEL6) >> SHIFT6) + which ; }
				if( tile & PIXEL5 ){ *( palnum +2 ) = ((tile & PIXEL5) >> SHIFT5) + which ; }
				if( tile & PIXEL4 ){ *( palnum +3 ) = ((tile & PIXEL4) >> SHIFT4) + which ; }
				if( tile & PIXEL3 ){ *( palnum +4 ) = ((tile & PIXEL3) >> SHIFT3) + which ; }
				if( tile & PIXEL2 ){ *( palnum +5 ) = ((tile & PIXEL2) >> SHIFT2) + which ; }
				if( tile & PIXEL1 ){ *( palnum +6 ) = ((tile & PIXEL1) >> SHIFT1) + which ; }
				if( tile & PIXEL0 ){ *( palnum +7 ) = ((tile & PIXEL0) >> SHIFT0) + which ; }
			//*/
			}
			else
			{
				which = ((which >> 13) & 0x0003) ;
				lay_tbl_jmp_A[which][ (unsigned char)(tile >> 24) ]( palnum ) ; palnum += 2 ;
				lay_tbl_jmp_A[which][ (unsigned char)(tile >> 16) ]( palnum ) ; palnum += 2 ;
				lay_tbl_jmp_A[which][ (unsigned char)(tile >>  8) ]( palnum ) ; palnum += 2 ;
				lay_tbl_jmp_A[which][ (unsigned char)(tile      ) ]( palnum ) ;
			}
		}
		else
		{
			if( (int)palnum % 2 )
			{
			//*
				which = ((which >> 9) & 0x0030) ;
				if( tile & PIXEL0 ){ *( palnum    ) = ((tile & PIXEL0) >> SHIFT0) + which ; }
				if( tile & PIXEL1 ){ *( palnum +1 ) = ((tile & PIXEL1) >> SHIFT1) + which ; }
				if( tile & PIXEL2 ){ *( palnum +2 ) = ((tile & PIXEL2) >> SHIFT2) + which ; }
				if( tile & PIXEL3 ){ *( palnum +3 ) = ((tile & PIXEL3) >> SHIFT3) + which ; }
				if( tile & PIXEL4 ){ *( palnum +4 ) = ((tile & PIXEL4) >> SHIFT4) + which ; }
				if( tile & PIXEL5 ){ *( palnum +5 ) = ((tile & PIXEL5) >> SHIFT5) + which ; }
				if( tile & PIXEL6 ){ *( palnum +6 ) = ((tile & PIXEL6) >> SHIFT6) + which ; }
				if( tile & PIXEL7 ){ *( palnum +7 ) = ((tile & PIXEL7) >> SHIFT7) + which ; }
			//*/
			}
			else
			{
				which = ((which >> 13) & 0x0003) ;
				lay_tbl_jmp_B[which][ (unsigned char)(tile      ) ]( palnum ) ; palnum += 2 ;
				lay_tbl_jmp_B[which][ (unsigned char)(tile >>  8) ]( palnum ) ; palnum += 2 ;
				lay_tbl_jmp_B[which][ (unsigned char)(tile >> 16) ]( palnum ) ; palnum += 2 ;
				lay_tbl_jmp_B[which][ (unsigned char)(tile >> 24) ]( palnum ) ;
			}
		}
	}
}

#else // _USE_LAY_TBL

inline void md_vdp::draw_lay_fast( unsigned long which, unsigned long tile, unsigned char* palnum )
{
	if( tile )
	{
		if( which & 0x0800 ) // x flipped
		{
			which = ((which >> 9) & 0x0030) ;
			if( tile & PIXEL7 ){ *( palnum ) = ((tile & PIXEL7) >> SHIFT7) + which ; } palnum++;
			if( tile & PIXEL6 ){ *( palnum ) = ((tile & PIXEL6) >> SHIFT6) + which ; } palnum++;
			if( tile & PIXEL5 ){ *( palnum ) = ((tile & PIXEL5) >> SHIFT5) + which ; } palnum++;
			if( tile & PIXEL4 ){ *( palnum ) = ((tile & PIXEL4) >> SHIFT4) + which ; } palnum++;
			if( tile & PIXEL3 ){ *( palnum ) = ((tile & PIXEL3) >> SHIFT3) + which ; } palnum++;
			if( tile & PIXEL2 ){ *( palnum ) = ((tile & PIXEL2) >> SHIFT2) + which ; } palnum++;
			if( tile & PIXEL1 ){ *( palnum ) = ((tile & PIXEL1) >> SHIFT1) + which ; } palnum++;
			if( tile & PIXEL0 ){ *( palnum ) = ((tile & PIXEL0) >> SHIFT0) + which ; }
		}
		else
		{
			which = ((which >> 9) & 0x0030) ;
			if( tile & PIXEL0 ){ *( palnum ) = ((tile & PIXEL0) >> SHIFT0) + which ; } palnum++;
			if( tile & PIXEL1 ){ *( palnum ) = ((tile & PIXEL1) >> SHIFT1) + which ; } palnum++;
			if( tile & PIXEL2 ){ *( palnum ) = ((tile & PIXEL2) >> SHIFT2) + which ; } palnum++;
			if( tile & PIXEL3 ){ *( palnum ) = ((tile & PIXEL3) >> SHIFT3) + which ; } palnum++;
			if( tile & PIXEL4 ){ *( palnum ) = ((tile & PIXEL4) >> SHIFT4) + which ; } palnum++;
			if( tile & PIXEL5 ){ *( palnum ) = ((tile & PIXEL5) >> SHIFT5) + which ; } palnum++;
			if( tile & PIXEL6 ){ *( palnum ) = ((tile & PIXEL6) >> SHIFT6) + which ; } palnum++;
			if( tile & PIXEL7 ){ *( palnum ) = ((tile & PIXEL7) >> SHIFT7) + which ; }
		}
	}
}

//*
inline void md_vdp::draw_lay_fast2( unsigned long which, unsigned long tile, unsigned char* palnum )
{
	if( tile )
	{
		if( tile & PIXEL7 ){ *( palnum ) = ((tile & PIXEL7) >> SHIFT7) + which ; } palnum++;
		if( tile & PIXEL6 ){ *( palnum ) = ((tile & PIXEL6) >> SHIFT6) + which ; } palnum++;
		if( tile & PIXEL5 ){ *( palnum ) = ((tile & PIXEL5) >> SHIFT5) + which ; } palnum++;
		if( tile & PIXEL4 ){ *( palnum ) = ((tile & PIXEL4) >> SHIFT4) + which ; } palnum++;
		if( tile & PIXEL3 ){ *( palnum ) = ((tile & PIXEL3) >> SHIFT3) + which ; } palnum++;
		if( tile & PIXEL2 ){ *( palnum ) = ((tile & PIXEL2) >> SHIFT2) + which ; } palnum++;
		if( tile & PIXEL1 ){ *( palnum ) = ((tile & PIXEL1) >> SHIFT1) + which ; } palnum++;
		if( tile & PIXEL0 ){ *( palnum ) = ((tile & PIXEL0) >> SHIFT0) + which ; }
	}
}

inline void md_vdp::draw_lay_fast3( unsigned long which, unsigned long tile, unsigned char* palnum )
{
	if( tile )
	{
		if( tile & PIXEL0 ){ *( palnum ) = ((tile & PIXEL0) >> SHIFT0) + which ; } palnum++;
		if( tile & PIXEL1 ){ *( palnum ) = ((tile & PIXEL1) >> SHIFT1) + which ; } palnum++;
		if( tile & PIXEL2 ){ *( palnum ) = ((tile & PIXEL2) >> SHIFT2) + which ; } palnum++;
		if( tile & PIXEL3 ){ *( palnum ) = ((tile & PIXEL3) >> SHIFT3) + which ; } palnum++;
		if( tile & PIXEL4 ){ *( palnum ) = ((tile & PIXEL4) >> SHIFT4) + which ; } palnum++;
		if( tile & PIXEL5 ){ *( palnum ) = ((tile & PIXEL5) >> SHIFT5) + which ; } palnum++;
		if( tile & PIXEL6 ){ *( palnum ) = ((tile & PIXEL6) >> SHIFT6) + which ; } palnum++;
		if( tile & PIXEL7 ){ *( palnum ) = ((tile & PIXEL7) >> SHIFT7) + which ; }
	}
}
//*/

#endif // _USE_LAY_TBL

#ifdef WIN32
static unsigned char draw_palnum[320+32] ;
#else
static unsigned char draw_palnum[320+32] __attribute__((aligned(64))) ;
#endif //WIN32


void md_vdp::draw_sprites_fast(int line, int front )
{
	unsigned which ;
	int tx, ty, x, y, xend, ysize, yoff, i;
	unsigned char *sprite;
	unsigned long tile ;
	int sh1, sh2 ;
	unsigned char *va, *vb ;
	unsigned char* palnum_t ;

	// interlace
	if(reg[12] & 2){ sh1 = 6 ; sh2 = 3 ; }else{ sh1 = 5 ; sh2 = 2 ; }

	// Sprites have to be in reverse order :P
	for(i = sprite_count - 1; i >= 0; --i)
	{
		sprite = sprite_base + (sprite_order[i] << 3);
		// Get the first tile
		which = get_word(sprite + 4);
		// Only do it if it's on the right priority
		if( (which >> 15) == front )
		{
			// Get the sprite's location
			x = (get_word(sprite + 6) & 0x1ff) ;
			//x -= 0x80;

			// Narrow mode?
			if( !(reg[12] & 1) ){ x += 32 - 0x80 ; }else{ x -= 0x80 ; }

			if( x < 320 )
			{
				y = get_word(sprite);

				// Interlace?
				if(reg[12] & 2){ y = (y & 0x3fe) >> 1; }else{ y &= 0x1ff; }

				y -= 0x80;
				yoff = (line - y);
				ysize = sprite[2] & 0x03 ;

				if( (0 <= yoff) && (yoff <= ((ysize<<3)+7)) )
				{
					xend = ((sprite[2] << 1) & 0x18) + x ;

					// Render if this sprite's on this line
					if( xend > -8 ) //&& x < 320 && yoff >= 0 && yoff <= (ysize<<3)+7)
					{
						// y flipped?
						if( which & 0x1000 )
						{
							va = (vram + (((yoff & 7)^7) << sh2));
							ty = (which & 0x7ff) - (yoff >> 3) + ysize ;
						}
						else
						{
							va = (vram + (((yoff & 7)  ) << sh2));
							ty = (which & 0x7ff) + (yoff >> 3)         ;
						}

						++ysize;

						// x flipped?
						if(which & 0x0800)
						{
							palnum_t = draw_palnum + 8 + xend ;

							which = ((which >> 9) & 0x0030) ;

							for( ; xend > 320 ; xend -= 8)
							{
								ty       += ysize ;
								palnum_t -= 8     ;
							}

							if( x < -7 ){ x = -7 ; }

							for(tx = xend; tx >= x ; tx -= 8)
							{
								tile = *(unsigned*)(va + ((ty & 0x7ff) << sh1) );

								draw_lay_fast2( which, tile, palnum_t ) ;

								ty       += ysize ;
								palnum_t -= 8     ;
							}
						} 
						else 
						{
							palnum_t = draw_palnum + 8 + x ;

							which = ((which >> 9) & 0x0030) ;
							
							if( xend > (320) ){ xend = 320 ; }

							for( ; x < -7 ; x += 8)
							{
								ty       += ysize ;
								palnum_t += 8     ;
							}

							for(tx = x; tx <= xend ; tx += 8)
							{
								tile = *(unsigned*)(va + ((ty & 0x7ff) << sh1) );

								draw_lay_fast3( which, tile, palnum_t ) ;

								ty       += ysize ;
								palnum_t += 8     ;
							}
						}
					}
				}
			}
		}
	}
}

void md_vdp::draw_window_fast (int line, int front )
{
	int size, xpos ;
	int x, y, w, start;
	int pl, add;
	int total_window;
	unsigned char *where;
	unsigned char *va, *vb ;
	unsigned long which, tile ;
	int sh1 ;
	unsigned char* palnum_t ;
	unsigned long size_m ;
	unsigned char* vpl ;

	// Set everything up
	y = line >> 3;
	total_window = (y < (reg[18]&0x1f)) ^ (reg[18] >> 7);

	// Wide or narrow
	size   = (reg[12] & 1)? 64 : 32;
	size_m = ((size-1)<<1) ;

	//pl = (reg[3] << 10) + ((y&0x3f)*size*2);
	vpl = vram + (reg[3] << 10) + ((y&0x3f)*size*2);

	// interlace
	if(reg[12] & 2)
	{ 
		sh1 = 6 ;
		va = ( vram + (((line&7)^7) << 3) ) ;
		vb = ( vram + (((line&7)  ) << 3) ) ;
	}
	else
	{ 
		sh1 = 5 ;
		va = ( vram + (((line&7)^7) << 2) ) ;
		vb = ( vram + (((line&7)  ) << 2) ) ;
	}

	// Wide(320) or narrow(256)?
	if(reg[12] & 1){ w = 40; start = -8; } else { w = 32; start = 24; }
	add = -2;

	palnum_t = draw_palnum + 8 + start ;

	if( !total_window )
	{
		xpos = ((reg[17]&0x1f) << 1) ;

		if(reg[17] & 0x80) 
		{
			add      += 2 * (xpos+1) ;
			palnum_t += 8 * (xpos+1) ;
		}
		else
		{
			w    = xpos ;
			xpos = -1   ;
		}
	}
	else
	{
		xpos = -1;
	}

	for( x=xpos; x < w ; ++x )
	{
		which = get_word( (unsigned char*)(vpl + (add & size_m)) );

		if( (which>>15) == front )
		{
			if( which & 0x1000 )
			{
				tile = *(unsigned*)(va + ((which & 0x7ff) << sh1) );
			}
			else
			{
				tile = *(unsigned*)(vb + ((which & 0x7ff) << sh1) );
			}

			draw_lay_fast( which, tile, palnum_t ) ;
		}
		add += 2; //where += Bpp_times8;
		palnum_t += 8 ;
	}
}

void md_vdp::draw_line_fast( int line )
{
	int xsize, ysize;
	int x, w, xstart, xpos ;
	unsigned char *scrolls0, *scrolls1;
	unsigned char *tiles0  , *tiles1  ;
	unsigned char *tile_line0, *tile_line1 ;
	int xoff0, xoff1, yoff0, yoff1, xoff_mask;
	int xoff0a, xoff1a ;
	int xscroll0, xscroll1, yscroll0, yscroll1;
	int sh1, sh2 ;
	int p0, shadow ;
	unsigned char *scan0, *scan1, *scan0a, *scan1a ;
	unsigned long which, tile ;
	unsigned char* palnum_t ;

	switch( reg[16] & 3 )
	{
	default :
	case 0 : xsize =  32 << 1 ; break ;
	case 1 : xsize =  64 << 1 ; break ;
	case 2 : xsize =  64 << 1 ; break ;
	case 3 : xsize = 128 << 1 ; break ;
	}

	switch( (reg[16] >> 4) & 3 )
	{
	default :
	case 0 : ysize =  32 ; break ;
	case 1 : ysize =  64 ; break ;
	case 2 : ysize =  64 ; break ;
	case 3 : ysize = 128 ; break ;
	}

	yscroll0 = get_word(vsram  );
	scrolls0 = vram + ((reg[13]<<10) & 0xfc00);
	tiles0   = vram +  (reg[ 2]<<10);
	yscroll1 = get_word(vsram+2);
	scrolls1 = vram + ((reg[13]<<10) & 0xfc00) + 2;
	tiles1   = vram +  (reg[ 4]<<13);

	switch( reg[12] & 6 )
	{
	// Interlace
	case 2 : 
	case 6 : 
		yscroll0 >>= 1; 
		yscroll1 >>= 1; 
		sh1 = 6 ; sh2 = 3 ;
		break ;

	// Interlace Double
	//case 6 : 
	//	sh1 = 6 ; sh2 = 3 ;
	//	break ;

	// Normal
	default :
		sh1 = 5 ; sh2 = 2 ;
		//sh1 = 6 ; sh2 = 3 ;
		break ;
	}

	// Offset for the line
	yscroll0 += line;
	yscroll1 += line;

	// Wide or narrow?
	if(reg[12] & 1){ w = 40; xstart = -8; } else { w = 32; xstart = 24; }

	// How do we horizontal offset?
	switch(reg[11] & 3)
	{
	case 1: // per tile
		scrolls0 += (line & 7) << 2;
		scrolls1 += (line & 7) << 2;
		break;
	case 2: // per tile
		scrolls0 += (line & ~7) << 2;
		scrolls1 += (line & ~7) << 2;
		break;
	case 3: // per line
		scrolls0 += line << 2;
		scrolls1 += line << 2;
		break;
	default:
		break;
	}


	xscroll0 = get_word(scrolls0);
	xscroll1 = get_word(scrolls1);

	xoff_mask = xsize - 1;

	yoff0 = (yscroll0>>3) & (ysize - 1);
	yoff1 = (yscroll1>>3) & (ysize - 1);
	xoff0 = ((-(xscroll0>>3) - 1)<<1) & xoff_mask;
	xoff1 = ((-(xscroll1>>3) - 1)<<1) & xoff_mask;
	tile_line0 = tiles0 + xsize * yoff0;
	tile_line1 = tiles1 + xsize * yoff1;
	scan0  = (vram + (( yscroll0 & 7)    << sh2) );
	scan1  = (vram + (( yscroll1 & 7)    << sh2) );
	scan0a = (vram + (((yscroll0 & 7)^7) << sh2) );
	scan1a = (vram + (((yscroll1 & 7)^7) << sh2) );

	xscroll0 = xstart + (xscroll0 & 7) ;
	xscroll1 = xstart + (xscroll1 & 7) ;
//	where0 = dest + ( xscroll0 ) * Bpp;
//	where1 = dest + ( xscroll1 ) * Bpp;

	//p0 = 1 ;
	if(reg[18] & 0x80) // Window goes down, plane 0 goes up! :)
	{
		p0 = ((line>>3) >= (reg[18]&0x1f)) ? 0 : 1 ;
	}
	else
	{
		p0 = ((line>>3) < (reg[18]&0x1f)) ? 0 : 1 ;
	}

	//*
	// Shadow / Hilight
	if( reg[12] & 8 )
	{
		shadow = 1 ;
	}
	else
	{
		shadow = 0 ;
	}
	//*/

	xpos = ((reg[17]&0x1f) << 1) ;
	if( xpos > 40 ){ xpos = 40 ; }

	which = reg[7] & 0x3F ; //backcolor
	which = (which << 24) | (which << 16) | (which << 8) | which ;
	for( x=0 ; x < sizeof(draw_palnum)/4 ; x++ ){ ((unsigned long*)draw_palnum)[x] = which ; }

//*
//#define _PLANE_A0_OFF
//#define _PLANE_A1_OFF
//#define _PLANE_B0_OFF
//#define _PLANE_B1_OFF
//#define _WND_0_OFF
//#define _WND_1_OFF
//#define _SPRITE_0_OFF
//#define _SPRITE_1_OFF
//*/

#ifndef _PLANE_A0_OFF
	//* Plane A - 0
	xoff1a = xoff1 ;
	palnum_t = draw_palnum + 8 + xscroll1 ;
	for( x=-1; x < w; x++ )
	{
		which = get_word(tile_line1 + xoff1a);

		if( !(which & 0x8000) )
		{
			if( which & 0x1000 ) // y flipped
			{
				tile = *(unsigned*)(scan1a + ((which & 0x7ff) << sh1) );
			}
			else
			{
				tile = *(unsigned*)(scan1  + ((which & 0x7ff) << sh1)  );
			}

			draw_lay_fast( which, tile, palnum_t ) ;
		}

		xoff1a = (xoff1a + 2) & xoff_mask ;
		palnum_t += 8 ;
	}
	//*/
#endif // _PLANE_A0_OFF

	if( p0 != 0 )
	{

#ifndef _PLANE_B0_OFF
		//* Plane B - 0
		xoff0a = xoff0 ;
		palnum_t = draw_palnum + 8 + xscroll0 ;
		if(reg[17] & 0x80) // Don't draw where the window will be
		{
			for( x = -1; x < xpos ; ++x )
			{
				which = get_word(tile_line0 + xoff0a);

				if( !(which & 0x8000) )
				{
					if( which & 0x1000 ) // y flipped
					{
						tile = *(unsigned*)(scan0a + ((which & 0x7ff) << sh1) ) ;
					}
					else
					{
						tile = *(unsigned*)(scan0  + ((which & 0x7ff) << sh1) ) ;
					}

					draw_lay_fast( which, tile, palnum_t ) ;
				}

				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8 ;
			}
		} 
		else 
		{
			xoff0a = (xoff0a + 2 * xpos) & xoff_mask ;
			palnum_t += 8 * xpos ;

			for(x = xpos -1; x < w; ++x)
			{
				which = get_word(tile_line0 + xoff0a);

				if( !(which & 0x8000) )
				{
					if( which & 0x1000 ) // y flipped
					{
						tile = *(unsigned*)(scan0a + ((which & 0x7ff) << sh1) ) ;
					}
					else
					{
						tile = *(unsigned*)(scan0  + ((which & 0x7ff) << sh1) ) ;
					}

					draw_lay_fast( which, tile, palnum_t ) ;
				}

				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8 ;
			}	
		}
		//*/
#endif // _PLANE_B0_OFF

		/*
		if( line < 32 )
		{
			draw_sprites_fast ( line , 0 );
			draw_window_fast  ( line , 0 );
		}
		else
		{
			draw_window_fast  ( line , 0 );
			draw_sprites_fast ( line , 0 );
		}
		*/
//*
#ifndef _WND_0_OFF
		draw_window_fast  ( line , 0 );
#endif // _WND_0_OFF

#ifndef _SPRITE_0_OFF
		draw_sprites_fast ( line , 0 );
#endif // _SPRITE_0_OFF
//*/
	}
	else
	{
		/*
		if( 1 )
		{
			draw_sprites_fast ( line , 1 );
			draw_window_fast  ( line , 1 );
		}
		else
		{
			draw_window_fast  ( line , 0 );
			draw_sprites_fast ( line , 0 );
		}
		//*/
//*
#ifndef _WND_0_OFF
		draw_window_fast  ( line , 0 );
#endif // _WND_0_OFF

#ifndef _SPRITE_0_OFF
		draw_sprites_fast ( line , 0 ); 
#endif // _SPRITE_0_OFF
//*/
	}

#ifndef _PLANE_A1_OFF
	//* Plane A - 1
	xoff1a = xoff1 ;
	palnum_t = draw_palnum + 8 + xscroll1 ;
	for( x=-1; x < w; x++ )
	{
		which = get_word(tile_line1 + xoff1a);

		if( (which & 0x8000) )
		{
			if( which & 0x1000 ) // y flipped
			{
				tile = *(unsigned*)(scan1a + ((which & 0x7ff) << sh1) );
			}
			else
			{
				tile = *(unsigned*)(scan1  + ((which & 0x7ff) << sh1)  );
			}

			draw_lay_fast( which, tile, palnum_t ) ;
		}

		xoff1a = (xoff1a + 2) & xoff_mask ;
		palnum_t += 8 ;
	}
	//*/
#endif // _PLANE_A1_OFF


	if( p0 != 0 )
	{
#ifndef _PLANE_B1_OFF
		//* Plane B - 1
		xoff0a = xoff0 ;
		palnum_t = draw_palnum + 8 + xscroll0 ;
		if(reg[17] & 0x80) // Don't draw where the window will be
		{
			for( x = -1; x < xpos ; ++x )
			{
				which = get_word(tile_line0 + xoff0a);

				if( (which & 0x8000) )
				{
					if( which & 0x1000 ) // y flipped
					{
						tile = *(unsigned*)(scan0a + ((which & 0x7ff) << sh1) ) ;
					}
					else
					{
						tile = *(unsigned*)(scan0  + ((which & 0x7ff) << sh1) ) ;
					}

					draw_lay_fast( which, tile, palnum_t ) ;
				}

				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8;
			}
		} 
		else 
		{
			xoff0a = (xoff0a + 2 * xpos) & xoff_mask ;
			palnum_t += 8 * xpos ;

			for(x = xpos -1; x < w; ++x)
			{
				which = get_word(tile_line0 + xoff0a);

				if( (which & 0x8000) )
				{
					if( which & 0x1000 ) // y flipped
					{
						tile = *(unsigned*)(scan0a + ((which & 0x7ff) << sh1) ) ;
					}
					else
					{
						tile = *(unsigned*)(scan0  + ((which & 0x7ff) << sh1) ) ;
					}

					draw_lay_fast( which, tile, palnum_t ) ;
				}

				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8 ;
			}
		}
		//*/
#endif // _PLANE_B1_OFF

/*
		if( line < 32 )
		{
			draw_sprites_fast ( line , 1 );
			draw_window_fast  ( line , 1 );
		}
		else
		{
			draw_window_fast  ( line , 1 );
			draw_sprites_fast ( line , 1 );
		}
*/

//*
#ifndef _WND_1_OFF
		draw_window_fast  ( line , 1 );
#endif // _WND_0_OFF

#ifndef _SPRITE_1_OFF
		draw_sprites_fast ( line , 1 ); 
#endif // _SPRITE_1_OFF
//*/
	}
	else
	{
		/*
		if( 1 )
		{
			//draw_window_fast  ( line , 0 );
			//draw_window_fast  ( line , 0 );
			draw_sprites_fast ( line , 0 );
			draw_window_fast  ( line , 0 );
			//draw_window_fast  ( line , 0 );
		}
		else
		{
			draw_window_fast  ( line , 1 );
			draw_sprites_fast ( line , 1 );
		}
		//*/

//*
#ifndef _WND_1_OFF
		draw_window_fast  ( line , 1 );
#endif // _WND_1_OFF

#ifndef _SPRITE_1_OFF
		draw_sprites_fast ( line , 1 ); 
#endif // _SPRITE_1_OFF
//*/
	}

	draw_pal2dest( xstart ) ;
}

void md_vdp::draw_cell0 ( unsigned char* d, unsigned char* t, unsigned char* scan1, unsigned char* scan1a, int sh1 )
{
	unsigned long which, tile ;
	
	which = get_word( t );

	if( !(which & 0x8000) )
	{
		if( which & 0x1000 ) // y flipped
		{
			tile = *(unsigned*)(scan1a + ((which & 0x7ff) << sh1) );
		}
		else
		{
			tile = *(unsigned*)(scan1  + ((which & 0x7ff) << sh1)  );
		}

		draw_lay_fast( which, tile, d ) ;
	}
}

void md_vdp::draw_cell1 ( unsigned char* d, unsigned char* t, unsigned char* scan1, unsigned char* scan1a, int sh1 )
{
	unsigned long which, tile ;
	
	which = get_word( t );

	if(  (which & 0x8000) )
	{
		if( which & 0x1000 ) // y flipped
		{
			tile = *(unsigned*)(scan1a + ((which & 0x7ff) << sh1) );
		}
		else
		{
			tile = *(unsigned*)(scan1  + ((which & 0x7ff) << sh1)  );
		}

		draw_lay_fast( which, tile, d ) ;
	}
}

void md_vdp::draw_line_fast_r( int line )
{
	int xsize, ysize;
	int x, w, xstart, xpos ;
	unsigned char *scrolls0, *scrolls1;
	unsigned char *tiles0  , *tiles1  ;
	int xoff0, xoff1, yoff0, yoff1, xoff_mask;
	int xoff0a, xoff1a ;
	int xscroll0, xscroll1, yscroll0, yscroll1;
	int sh1, sh2 ;
	int p0, shadow ;
	unsigned long which, tile ;
	unsigned char* palnum_t ;
	unsigned char *scan0[21], *scan0a[21], *tile_line0[21] ;
	unsigned char *scan1[21], *scan1a[21], *tile_line1[21] ;

	switch( reg[16] & 3 )
	{
	default :
	case 0 : xsize =  32 << 1 ; break ;
	case 1 : xsize =  64 << 1 ; break ;
	case 2 : xsize =  64 << 1 ; break ;
	case 3 : xsize = 128 << 1 ; break ;
	}

	switch( (reg[16] >> 4) & 3 )
	{
	default :
	case 0 : ysize =  32 ; break ;
	case 1 : ysize =  64 ; break ;
	case 2 : ysize =  64 ; break ;
	case 3 : ysize = 128 ; break ;
	}

	scrolls0 = vram + ((reg[13]<<10) & 0xfc00);
	tiles0   = vram +  (reg[ 2]<<10);
	scrolls1 = vram + ((reg[13]<<10) & 0xfc00) + 2;
	tiles1   = vram +  (reg[ 4]<<13);

	// Wide or narrow?
	if(reg[12] & 1){ w = 40; xstart = -8; } else { w = 32; xstart = 24; }

	// How do we horizontal offset?
	switch(reg[11] & 3)
	{
	case 1: // per tile
		scrolls0 += (line & 7) << 2;
		scrolls1 += (line & 7) << 2;
		break;
	case 2: // per tile
		scrolls0 += (line & ~7) << 2;
		scrolls1 += (line & ~7) << 2;
		break;
	case 3: // per line
		scrolls0 += line << 2;
		scrolls1 += line << 2;
		break;
	default:
		break;
	}

	switch( reg[12] & 6 )
	{
	// Interlace
	case 2 : 
	case 6 : 
		sh1 = 6 ; sh2 = 3 ;
		break ;

	// Interlace Double
	//case 6 : 
	//	sh1 = 6 ; sh2 = 3 ;
	//	break ;

	// Normal
	default :
		sh1 = 5 ; sh2 = 2 ;
		//sh1 = 6 ; sh2 = 3 ;
		break ;
	}


	for( x=0; x <= 40 / 2; x++ )
	{
		//if( w == 32 )
		{
			if( x > 0 )
			{
				yscroll0 = get_word( vsram + (x-1)*4    );
				yscroll1 = get_word( vsram + (x-1)*4 +2 );
			}
			else
			{
				yscroll0 = get_word( vsram + x*4    );
				yscroll1 = get_word( vsram + x*4 +2 );
			}
		}
		/*
		else
		{
			if( x > 0 )
			{
				yscroll0 = get_word( vsram + (x+1)*4    );
				yscroll1 = get_word( vsram + (x+1)*4 +2 );
			}
			else
			{
				yscroll0 = get_word( vsram + x*4    );
				yscroll1 = get_word( vsram + x*4 +2 );
			}
		}
		//*/

		if( reg[12] & 2 )
		{
			yscroll0 >>= 1; 
			yscroll1 >>= 1; 
		}

		// Offset for the line
		yscroll0 += line;
		yscroll1 += line;

		yoff0 = (yscroll0>>3) & (ysize - 1);
		yoff1 = (yscroll1>>3) & (ysize - 1);

		tile_line0[x] = tiles0 + xsize * yoff0 ;
		tile_line1[x] = tiles1 + xsize * yoff1 ;
		scan0     [x] = ( vram + (( yscroll0 & 7)    << sh2) );
		scan1     [x] = ( vram + (( yscroll1 & 7)    << sh2) );
		scan0a    [x] = ( vram + (((yscroll0 & 7)^7) << sh2) );
		scan1a    [x] = ( vram + (((yscroll1 & 7)^7) << sh2) );
	}

	xscroll0  = get_word( scrolls0 );
	xscroll1  = get_word( scrolls1 );
	xoff_mask = xsize - 1;
	xoff0     = ((-(xscroll0>>3) - 1)<<1) & xoff_mask;
	xoff1     = ((-(xscroll1>>3) - 1)<<1) & xoff_mask;
	xscroll0  = xstart + (xscroll0 & 7) ;
	xscroll1  = xstart + (xscroll1 & 7) ;

	//p0 = 1 ;
	if(reg[18] & 0x80) // Window goes down, plane 0 goes up! :)
	{
		p0 = ((line>>3) >= (reg[18]&0x1f)) ? 0 : 1 ;
	}
	else
	{
		p0 = ((line>>3) < (reg[18]&0x1f)) ? 0 : 1 ;
	}

	//*
	// Shadow / Hilight
	if( reg[12] & 8 )
	{
		shadow = 1 ;
	}
	else
	{
		shadow = 0 ;
	}
	//*/

	xpos = ((reg[17]&0x1f) << 1) ;
	if( xpos > 40 ){ xpos = 40 ; }

	which = reg[7] & 0x3F ; //backcolor
	which = (which << 24) | (which << 16) | (which << 8) | which ;
	for( x=0 ; x < sizeof(draw_palnum)/4 ; x++ ){ ((unsigned long*)draw_palnum)[x] = which ; }

//*
//#define _PLANE_A0_OFF
//#define _PLANE_A1_OFF
//#define _PLANE_B0_OFF
//#define _PLANE_B1_OFF
//#define _WND_0_OFF
//#define _WND_1_OFF
//#define _SPRITE_0_OFF
//#define _SPRITE_1_OFF
//*/

#ifndef _PLANE_A0_OFF
	//* Plane A - 0
	xoff1a = xoff1 ;
	palnum_t = draw_palnum + 8 + xscroll1 ;

	for( x=0; x <= w / 2 ; x++ )
	{
		draw_cell0( palnum_t, tile_line1[x] + xoff1a, scan1[x], scan1a[x], sh1 ) ;
		xoff1a = (xoff1a + 2) & xoff_mask ;
		palnum_t += 8 ;

		draw_cell0( palnum_t, tile_line1[x] + xoff1a, scan1[x], scan1a[x], sh1 ) ;
		xoff1a = (xoff1a + 2) & xoff_mask ;
		palnum_t += 8 ;
	}
	//*/
#endif // _PLANE_A0_OFF


	if( p0 != 0 )
	{

#ifndef _PLANE_B0_OFF
		//* Plane B - 0
		xoff0a = xoff0 ;
		palnum_t = draw_palnum + 8 + xscroll0 ;
		if(reg[17] & 0x80) // Don't draw where the window will be
		{
			for( x=0; x <= xpos / 2 ; x++ )
			{
				draw_cell0( palnum_t, tile_line0[x] + xoff0a, scan0[x], scan0a[x], sh1 ) ;
				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8 ;

				draw_cell0( palnum_t, tile_line0[x] + xoff0a, scan0[x], scan0a[x], sh1 ) ;
				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8 ;
			}
		} 
		else 
		{
			xoff0a = (xoff0a + 2 * xpos) & xoff_mask ;
			palnum_t += 8 * xpos ;

			for( x=0 ; x <= (w - xpos)/2 ; x++ )
			{
				draw_cell0( palnum_t, tile_line0[x] + xoff0a, scan0[x], scan0a[x], sh1 ) ;
				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8 ;

				draw_cell0( palnum_t, tile_line0[x] + xoff0a, scan0[x], scan0a[x], sh1 ) ;
				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8 ;
			}	
		}
		//*/
#endif // _PLANE_B0_OFF
		
#ifndef _WND_0_OFF
		draw_window_fast  ( line , 0 );
#endif // _WND_0_OFF

#ifndef _SPRITE_0_OFF
		draw_sprites_fast ( line , 0 );
#endif // _SPRITE_0_OFF
	}
	else
	{

#ifndef _WND_0_OFF
		draw_window_fast  ( line , 0 );
#endif // _WND_0_OFF

#ifndef _SPRITE_0_OFF
		draw_sprites_fast ( line , 0 ); 
#endif // _SPRITE_0_OFF
	}

#ifndef _PLANE_A1_OFF
	//* Plane A - 1
	xoff1a = xoff1 ;
	palnum_t = draw_palnum + 8 + xscroll1 ;
	for( x = 0; x <= w / 2; x++ )
	{
		draw_cell1( palnum_t, tile_line1[x] + xoff1a, scan1[x], scan1a[x], sh1 ) ;
		xoff1a = (xoff1a + 2) & xoff_mask ;
		palnum_t += 8 ;

		draw_cell1( palnum_t, tile_line1[x] + xoff1a, scan1[x], scan1a[x], sh1 ) ;
		xoff1a = (xoff1a + 2) & xoff_mask ;
		palnum_t += 8 ;
	}
	//*/
#endif // _PLANE_A1_OFF


	if( p0 != 0 )
	{
#ifndef _PLANE_B1_OFF
		//* Plane B - 1
		xoff0a = xoff0 ;
		palnum_t = draw_palnum + 8 + xscroll0 ;
		if(reg[17] & 0x80) // Don't draw where the window will be
		{
			for( x=0; x <= xpos / 2 ; x++ )
			{
				draw_cell1( palnum_t, tile_line0[x] + xoff0a, scan0[x], scan0a[x], sh1 ) ;
				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8;

				draw_cell1( palnum_t, tile_line0[x] + xoff0a, scan0[x], scan0a[x], sh1 ) ;
				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8;
			}
		} 
		else 
		{
			xoff0a = (xoff0a + 2 * xpos) & xoff_mask ;
			palnum_t += 8 * xpos ;

			for( x=0 ; x <= (w - xpos)/2; x++ )
			{
				draw_cell1( palnum_t, tile_line0[x] + xoff0a, scan0[x], scan0a[x], sh1 ) ;
				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8 ;

				draw_cell1( palnum_t, tile_line0[x] + xoff0a, scan0[x], scan0a[x], sh1 ) ;
				xoff0a = (xoff0a + 2) & xoff_mask ;
				palnum_t += 8 ;
			}
		}
		//*/
#endif // _PLANE_B1_OFF

#ifndef _WND_1_OFF
		draw_window_fast  ( line , 1 );
#endif // _WND_0_OFF

#ifndef _SPRITE_1_OFF
		draw_sprites_fast ( line , 1 ); 
#endif // _SPRITE_1_OFF
	}
	else
	{
#ifndef _WND_1_OFF
		draw_window_fast  ( line , 1 );
#endif // _WND_1_OFF

#ifndef _SPRITE_1_OFF
		draw_sprites_fast ( line , 1 ); 
#endif // _SPRITE_1_OFF
	}

	draw_pal2dest( xstart ) ;
}

void md_vdp::draw_pal2dest ( int xstart )
{
	int x ;
	unsigned char* dest_t ;
	unsigned long  which  ;

	//*
	if( xstart > -8 )
	{
		/*
		dest_t = dest ;
		for( x=0; x < 8 ; x++ )
		{ 
			*((unsigned short*)dest_t) = 0 ; dest_t += 2 ;
			*((unsigned short*)dest_t) = 0 ; dest_t += 2 ;
			*((unsigned short*)dest_t) = 0 ; dest_t += 2 ;
			*((unsigned short*)dest_t) = 0 ; dest_t += 2 ;
		}

		dest_t = dest + ( 40 - 4 ) * 8 * Bpp ;
		for( x=0; x < 8 ; x++ )
		{ 
			*((unsigned short*)dest_t) = 0 ; dest_t += 2 ;
			*((unsigned short*)dest_t) = 0 ; dest_t += 2 ;
			*((unsigned short*)dest_t) = 0 ; dest_t += 2 ;
			*((unsigned short*)dest_t) = 0 ; dest_t += 2 ;
		}
		//*/
		dest_t = dest ;
		for( x=0; x < 4 ; x++ )
		{ 
			*((unsigned long*)dest_t) = 0 ; dest_t += 4 ;
			*((unsigned long*)dest_t) = 0 ; dest_t += 4 ;
			*((unsigned long*)dest_t) = 0 ; dest_t += 4 ;
			*((unsigned long*)dest_t) = 0 ; dest_t += 4 ;
		}

		dest_t = dest + ( 40 - 4 ) * 8 * Bpp ;
		for( x=0; x < 4 ; x++ )
		{ 
			*((unsigned long*)dest_t) = 0 ; dest_t += 4 ;
			*((unsigned long*)dest_t) = 0 ; dest_t += 4 ;
			*((unsigned long*)dest_t) = 0 ; dest_t += 4 ;
			*((unsigned long*)dest_t) = 0 ; dest_t += 4 ;
		}

		//*
		dest_t = dest + (8 * 4 * Bpp) ;
		for( x=24+16 ; x < (32 * 8 + (24+16)) ; x+=4 )
		{ 
			which = *((unsigned long*)&draw_palnum[x]) ;
			*((unsigned short*)dest_t) = highpal[ (unsigned char)(which >>  0) ] ; dest_t += 2 ;
			*((unsigned short*)dest_t) = highpal[ (unsigned char)(which >>  8) ] ; dest_t += 2 ;
			*((unsigned short*)dest_t) = highpal[ (unsigned char)(which >> 16) ] ; dest_t += 2 ;
			*((unsigned short*)dest_t) = highpal[ (unsigned char)(which >> 24) ] ; dest_t += 2 ;
		}
		//*/
	}
	else
	//*/
	{
		dest_t = dest ;
		for( x=-8+16 ; x < (40 * 8 + (-8+16)) ; x+=4 )
		{ 
			which = *((unsigned long*)&draw_palnum[x]) ;
			*((unsigned short*)dest_t) = highpal[ (unsigned char)(which >>  0) ] ; dest_t += 2 ;
			*((unsigned short*)dest_t) = highpal[ (unsigned char)(which >>  8) ] ; dest_t += 2 ;
			*((unsigned short*)dest_t) = highpal[ (unsigned char)(which >> 16) ] ; dest_t += 2 ;
			*((unsigned short*)dest_t) = highpal[ (unsigned char)(which >> 24) ] ; dest_t += 2 ;
		}
	}
}

// The main interface function, to generate a scanline
void md_vdp::draw_scanline(struct bmap *bits, int line)
{
  unsigned short *ptr, i;
  // Set the destination in the bmap
  dest = bits->data + (bits->pitch * (line + 8) ) ; //+ 16);
  // If bytes per pixel hasn't yet been set, do it

	Bpp = 2;

  // If the palette's been changed, update it
  if(dirt[0x34] & 2)
    {
      ptr = highpal;
	  for(i = 0; i < 128; i += 2)
	    *ptr++ = ((cram[i+1]&0x0e) << 1 ) |
	    	     ((cram[i+1]&0xe0) << 2 ) |
				 ((cram[i]  &0x0e) << 11);

      // Clean up the dirt
      dirt[0x34] &= ~2;
      pal_dirty = 1;
    }

  // Render the screen if it's turned on
  if(reg[1] & 0x40)
    {
		// Recalculate the sprite order, if it's dirty
		if((dirt[0x30] & 0x20) || (dirt[0x34] & 1))
		{
			unsigned next = 0;
			// Find the sprite base in VRAM
			sprite_base = vram + (reg[5]<<9);
			// Order the sprites
			sprite_count = sprite_order[0] = 0;
			do {
				next = sprite_base[(next << 3) + 3];
				sprite_order[++sprite_count] = next;
				// No more than 256 sprites/line, a reasonable limit ;)
			//} while (next && sprite_count < 0x100);
			} while (next && sprite_count < 0x50);
			// Clean up the dirt
			dirt[0x30] &= ~0x20; dirt[0x34] &= ~1;
		}

		if( reg[11] & 4 )
		{
			draw_line_fast_r( line ) ;
		}
		else
		{
			draw_line_fast( line ) ;
		}
	} 
	else 
	{
      // The display is off, paint it black
      // Do it a dword at a time
      unsigned *destl = (unsigned*)dest ;
      for(i = 0; i < (80 * Bpp); ++i) destl[i] = 0 ;
    }
}

#else ///////////////////////////////////////////////////////////////////////////////


// The main interface function, to generate a scanline
void md_vdp::draw_scanline(struct bmap *bits, int line)
{
  unsigned short *ptr, i;
  // Set the destination in the bmap
  dest = bits->data + (bits->pitch * (line + 8) + 16);
  // If bytes per pixel hasn't yet been set, do it

  if(!Bpp)
    {
           if(bits->bpp <= 8)  Bpp = 1;
      else if(bits->bpp <= 16) Bpp = 2;
      else if(bits->bpp <= 24) Bpp = 3;
      else		       Bpp = 4;
      Bpp_times8 = Bpp << 3; // used for tile blitting
#ifdef ASM_TILES
      _asm_tiles_init(vram, reg, highpal); // pass these values to the asm tiles
#endif
    }

  // If the palette's been changed, update it
  if(dirt[0x34] & 2)
    {
      ptr = highpal;
      // What color depth are we?
      switch(bits->bpp)
        {
	case 32:
	case 24:
	  for(i = 0; i < 128; i += 2)
	    *ptr++ = ((cram[i+1]&0x0e) << 20) |
	    	     ((cram[i+1]&0xe0) << 8 ) |
		     ((cram[i]  &0x0e) << 4 );
	  break;
	case 16:
	  for(i = 0; i < 128; i += 2)
	    *ptr++ = ((cram[i+1]&0x0e) << 12) |
	    	     ((cram[i+1]&0xe0) << 3 ) |
		     ((cram[i]  &0x0e) << 1 );
	  break;
	case 15:
	  for(i = 0; i < 128; i += 2)
	    *ptr++ = ((cram[i+1]&0x0e) << 11) |
	    	     ((cram[i+1]&0xe0) << 2 ) |
			((cram[i]  &0x0e) << 1 );
	  break;
	case 8:
	default:
	  // Let the hardware palette sort it out :P
	  for(i = 0; i < 64; ++i) *ptr++ = i;
	}
      // Clean up the dirt
      dirt[0x34] &= ~2;
      pal_dirty = 1;
    }

  // Render the screen if it's turned on
  if(reg[1] & 0x40)
    {
		// Recalculate the sprite order, if it's dirty
		if((dirt[0x30] & 0x20) || (dirt[0x34] & 1))
		{
			unsigned next = 0;
			// Find the sprite base in VRAM
			sprite_base = vram + (reg[5]<<9);
			// Order the sprites
			sprite_count = sprite_order[0] = 0;
			do {
				next = sprite_base[(next << 3) + 3];
				sprite_order[++sprite_count] = next;
				// No more than 256 sprites/line, a reasonable limit ;)
			} while (next && sprite_count < 0x100);
			// Clean up the dirt
			dirt[0x30] &= ~0x20; dirt[0x34] &= ~1;
		}

      // Draw, from the bottom up
      // Low priority
      draw_plane_back1(line);
      draw_plane_back0(line);
      draw_window(line, 0);
      draw_sprites(line, 0);
      // High priority
      draw_plane_front1(line);
      draw_plane_front0(line);
      draw_window(line, 1);
      draw_sprites(line, 1);
    } else {
      // The display is off, paint it black
      // Do it a dword at a time
      unsigned *destl = (unsigned*)dest;
      for(i = 0; i < (80 * Bpp); ++i) destl[i] = 0;
    }

  // If we're in narrow (256) mode, cut off the messy edges
  if(!(reg[12] & 1))
    {
      unsigned *destl = (unsigned*)dest;
      for(i = 0; i < Bpp_times8; ++i)
        destl[i] = destl[i + (72 * Bpp)] = 0;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////
#endif //_PSP10
////////////////////////////////////////////////////////////////////////////////////////////
