// DGen v1.10+
// Megadrive C++ module saving and loading

#include <stdio.h>
#include <string.h>
#include "md.h"

/*
gs0 genecyst save file INfo

GST\0 to start with

80-9f = d0-d7 almost certain
a0-bf = a0-a7 almost certain
c8    = pc    fairly certain
d0    = sr    fairly certain


112 Start of cram len 0x80
192 Start of vsram len 0x50
1e2-474 UNKNOWN sound info?
Start of z80 ram at 474 (they store 2000)
Start of RAM at 2478 almost certain (BYTE SWAPPED)
Start of VRAM at 12478
end of VRAM
*/

#ifdef _PSP10

inline void memcpy4byte( void* dst, void* src )
{
	((unsigned char*)dst)[0] = ((unsigned char*)src)[0] ;
	((unsigned char*)dst)[1] = ((unsigned char*)src)[1] ;
	((unsigned char*)dst)[2] = ((unsigned char*)src)[2] ;
	((unsigned char*)dst)[3] = ((unsigned char*)src)[3] ;
}

int md::get_state( unsigned char* buff )
{
	int i, t ;
	int ret = 0 ;

    for (i=0;i<8;i++)
	{ 
		t=m68k_peek_dr(i); 
		memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
	}

    for (i=0;i<8;i++)
	{ 
		t=m68k_peek_ar(i);
		memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
	}

    t=m68k_peek_pc      (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
//    t=m68k_peek_ppc     (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_sr      (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_ir      (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_t1_flag (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_t0_flag (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_s_flag  (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_m_flag  (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_int_mask(); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_x_flag  (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_n_flag  (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_z_flag  (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_v_flag  (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_c_flag  (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_usp     (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_isp     (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;
    t=m68k_peek_msp     (); memcpy4byte( &buff[ret], &t ) ; ret += 4 ;

	memcpy( &buff[ret], ram      , 0x10000 ) ; ret += 0x10000 ;
	memcpy( &buff[ret], vdp.reg  , 0x00020 ) ; ret += 0x00020 ;
	memcpy( &buff[ret], vdp.vsram, 0x00050 ) ; ret += 0x00050 ;
	memcpy( &buff[ret], vdp.cram , 0x00080 ) ; ret += 0x00080 ;
	memcpy( &buff[ret], vdp.vram , 0x10000 ) ; ret += 0x10000 ;
	memcpy( &buff[ret], z80ram   , 0x02000 ) ; ret += 0x02000 ;

	memcpy( &buff[ret], &z80 , sizeof(z80) ) ; ret += sizeof(z80) ;
//	memcpy4byte( &buff[ret], &z80_int_pending  ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &z80_bank68k      ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &z80_online       ) ; ret += 4 ;
//	memcpy4byte( &buff[ret], &z80_extra_cycles ) ; ret += 4 ;

//	memcpy( &buff[ret], dac_data , sizeof(dac_data) ) ; ret += sizeof(dac_data) ;
	ret += sizeof( int[0x138] ) ; // * 0x138 ;
	memcpy4byte( &buff[ret], &dac_enabled      ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &dac_last         ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &pal              ) ; ret += 4 ;

	memcpy4byte( &buff[ret], &aoo3_toggle      ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &aoo3_six         ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &aoo3_six_timeout ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &aoo5_toggle      ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &aoo5_six         ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &aoo5_six_timeout ) ; ret += 4 ;

	memcpy4byte( &buff[ret], &coo_waiting      ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &coo_cmd          ) ; ret += 4 ;

	memcpy4byte( &buff[ret], &vdp.rw_mode      ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &vdp.rw_addr      ) ; ret += 4 ;
	memcpy4byte( &buff[ret], &vdp.rw_dma       ) ; ret += 4 ;

	ret += SN76496_get_state( &buff[ret] ) ;

	memcpy( &buff[ret], fm_sel        , sizeof(fm_sel       ) ) ; ret += sizeof(fm_sel       ) ;
	memcpy( &buff[ret], fm_tover      , sizeof(fm_tover     ) ) ; ret += sizeof(fm_tover     ) ;
	memcpy( &buff[ret], ras_fm_ticker , sizeof(ras_fm_ticker) ) ; ret += sizeof(ras_fm_ticker) ;
	memcpy( &buff[ret], fm_reg        , sizeof(fm_reg       ) ) ; ret += sizeof(fm_reg       ) ;

	ret += YM2612GetState( &buff[ret] ) ;

	return ret ;
}

int md::set_state( unsigned char* buff, int buff_len, int ver )
{
	int i, t ;
	int ret = 0 ;
	struct mz80context mz80ct ;

    for (i=0;i<8;i++)
	{ 
		memcpy4byte( &t, &buff[ret] ) ; ret += 4 ;
		m68k_poke_dr(i,t); 
	}

    for (i=0;i<8;i++)
	{ 
		memcpy4byte( &t, &buff[ret] ) ; ret += 4 ;
		m68k_poke_ar(i,t);
	}

	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_pc      (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_sr      (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_ir      (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_t1_flag (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_t0_flag (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_s_flag  (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_m_flag  (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_int_mask(t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_x_flag  (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_n_flag  (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_z_flag  (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_v_flag  (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_c_flag  (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_usp     (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_isp     (t);
	memcpy4byte( &t, &buff[ret] ) ; ret += 4 ; m68k_poke_msp     (t);

	memcpy( ram      , &buff[ret], 0x10000 ) ; ret += 0x10000 ;
	memcpy( vdp.reg  , &buff[ret], 0x00020 ) ; ret += 0x00020 ;
	memcpy( vdp.vsram, &buff[ret], 0x00050 ) ; ret += 0x00050 ;
	memcpy( vdp.cram , &buff[ret], 0x00080 ) ; ret += 0x00080 ;
	memcpy( vdp.vram , &buff[ret], 0x10000 ) ; ret += 0x10000 ;
	memcpy( z80ram   , &buff[ret], 0x02000 ) ; ret += 0x02000 ;

	memcpy( &mz80ct , &buff[ret], sizeof(z80) ) ; ret += sizeof(z80) ;
//	memcpy4byte( &z80_int_pending , &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &z80_bank68k     , &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &z80_online      , &buff[ret] ) ; ret += 4 ;
//	memcpy4byte( &z80_extra_cycles, &buff[ret] ) ; ret += 4 ;

//	memcpy( dac_data , &buff[ret], sizeof(dac_data) ) ; ret += sizeof(dac_data) ;
	ret += sizeof( int[0x138] ) ; // * 0x138 ;
	memcpy4byte( &dac_enabled     , &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &dac_last        , &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &pal             , &buff[ret] ) ; ret += 4 ;

	memcpy4byte( &aoo3_toggle     , &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &aoo3_six        , &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &aoo3_six_timeout, &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &aoo5_toggle     , &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &aoo5_six        , &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &aoo5_six_timeout, &buff[ret] ) ; ret += 4 ;

	memcpy4byte( &coo_waiting     , &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &coo_cmd         , &buff[ret] ) ; ret += 4 ;

	memcpy4byte( &vdp.rw_mode     , &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &vdp.rw_addr     , &buff[ret] ) ; ret += 4 ;
	memcpy4byte( &vdp.rw_dma      , &buff[ret] ) ; ret += 4 ;

	ret += SN76496_set_state( &buff[ret], buff_len - ret ) ;

	memcpy( fm_sel        , &buff[ret], sizeof(fm_sel       ) ) ; ret += sizeof(fm_sel       ) ;
	memcpy( fm_tover      , &buff[ret], sizeof(fm_tover     ) ) ; ret += sizeof(fm_tover     ) ;
	memcpy( ras_fm_ticker , &buff[ret], sizeof(ras_fm_ticker) ) ; ret += sizeof(ras_fm_ticker) ;
	memcpy( fm_reg        , &buff[ret], sizeof(fm_reg       ) ) ; ret += sizeof(fm_reg       ) ;

	if( ver > 1 )
	{
		ret += YM2612SetState( &buff[ret], buff_len - ret ) ;
	}
	flush_fm_to_mame() ;
	//YM2612SetRegs( (short*)&fm_reg[0][0] ) ;

	z80.z80clockticks    = mz80ct.z80clockticks     ;
	z80.z80iff           = mz80ct.z80iff            ;
	z80.z80interruptMode = mz80ct.z80interruptMode  ;
	z80.z80halted        = mz80ct.z80halted         ;
	z80.z80af            = mz80ct.z80af             ;
	z80.z80bc            = mz80ct.z80bc             ;
	z80.z80de            = mz80ct.z80de             ;
	z80.z80hl            = mz80ct.z80hl             ;
	z80.z80afprime       = mz80ct.z80afprime        ;
	z80.z80bcprime       = mz80ct.z80bcprime        ;
	z80.z80deprime       = mz80ct.z80deprime        ;
	z80.z80hlprime       = mz80ct.z80hlprime        ;
	z80.z80ix            = mz80ct.z80ix             ;
	z80.z80iy            = mz80ct.z80iy             ;
	z80.z80sp            = mz80ct.z80sp             ;
	z80.z80pc            = mz80ct.z80pc             ;
	z80.z80nmiAddr       = mz80ct.z80nmiAddr        ;
	z80.z80intAddr       = mz80ct.z80intAddr        ;
	z80.z80rCounter      = mz80ct.z80rCounter       ;
	z80.z80i             = mz80ct.z80i              ;
	z80.z80r             = mz80ct.z80r              ;
	z80.z80intPending    = mz80ct.z80intPending     ;

	memset(vdp.dirt,0xff,0x35); // mark everything as changed

	return ret ;
}


#else  //_PSP10

// NB - for load and save you don't need to use star_mz80_on/off
// Because stars/mz80 isn't actually used
#define fput(x,y) { if (fwrite(x,1,y,hand)!=y) goto save_error; }
#define fget(x,y) if (fread(x,1,y,hand)!=y) goto load_error;

extern int byteswap_memory(unsigned char *start,int len);

int md::import_gst(FILE *hand)
{
#ifdef COMPILE_WITH_STAR
  if (cpu_emu==0)
  {
    fseek(hand,0x80,SEEK_SET);
    fget(cpu.dreg,8*4);
    fget(cpu.areg,8*4);
    fseek(hand,0xc8,SEEK_SET);
    fget(&cpu.pc,4);
    fseek(hand,0xd0,SEEK_SET);
    fget(&cpu.sr,4);
  }
#endif

#ifdef COMPILE_WITH_MUSA
  if (cpu_emu==1)
  {
    int i,t;
    fseek(hand,0x80,SEEK_SET);
    for (i=0;i<8;i++)
    { fget(&t,4); m68k_poke_dr(i,t); }
    for (i=0;i<8;i++)
    { fget(&t,4); m68k_poke_ar(i,t); }

    fseek(hand,0xc8,SEEK_SET);
    fget(&t,4); m68k_poke_pc(t);

    fseek(hand,0xd0,SEEK_SET);
    fget(&t,4); m68k_poke_sr(t);
  }
#endif

  fseek(hand,0xfa,SEEK_SET);
  fget(vdp.reg,0x18);

  fseek(hand,0x112,SEEK_SET);
  fget(vdp.cram ,0x00080);
  byteswap_memory(vdp.cram,0x80);
  fget(vdp.vsram,0x00050);
  byteswap_memory(vdp.vsram,0x50);

  fseek(hand,0x474,SEEK_SET);
  fget(z80ram,   0x02000);

  fseek(hand,0x2478,SEEK_SET);
  fget(ram,      0x10000);
  byteswap_memory(ram,0x10000);

  fget(vdp.vram ,0x10000);

  memset(vdp.dirt,0xff,0x35); // mark everything as changed

  return 0;
load_error:
  return 1;
}


int md::export_gst(FILE *hand)
{
  int i;
  static unsigned char gst_head[0x80]=
  {
       0x47,0x53,0x54,0,0,0,0xe0,0x40

  //     00 00 00 00 00 00 00 00 00 00 00 00 00 21 80 fa   <.............!..>
  };
  unsigned char *zeros=gst_head+0x40; // 0x40 zeros

  fseek(hand,0x00,SEEK_SET);
  // Make file size 0x22478 with zeros
  for (i=0;i<0x22440;i+=0x40) fput(zeros,0x40);
  fput(zeros,0x38);

  fseek(hand,0x00,SEEK_SET);
  fput(gst_head,0x80);

#ifdef COMPILE_WITH_STAR
  if (cpu_emu==0)
  {
    fseek(hand,0x80,SEEK_SET);
    fput(cpu.dreg,8*4);
    fput(cpu.areg,8*4);
    fseek(hand,0xc8,SEEK_SET);
    fput(&cpu.pc,4);
    fseek(hand,0xd0,SEEK_SET);
    fput(&cpu.sr,4);
  }
#endif
#ifdef COMPILE_WITH_MUSA
  if (cpu_emu==1)
  {
    int i,t;
    fseek(hand,0x80,SEEK_SET);
    for (i=0;i<8;i++)
    { t=m68k_peek_dr(i); fput(&t,4);}
    for (i=0;i<8;i++)
    { t=m68k_peek_ar(i); fput(&t,4);}

    fseek(hand,0xc8,SEEK_SET);
    t=m68k_peek_pc(); fput(&t,4);

    fseek(hand,0xd0,SEEK_SET);
    t=m68k_peek_sr(); fput(&t,4);
  }
#endif

  fseek(hand,0xfa,SEEK_SET);
  fput(vdp.reg,0x18);

  fseek(hand,0x112,SEEK_SET);
  byteswap_memory(vdp.cram,0x80);
  fput(vdp.cram ,0x00080);
  byteswap_memory(vdp.cram,0x80);
  byteswap_memory(vdp.vsram,0x50);
  fput(vdp.vsram,0x00050);
  byteswap_memory(vdp.vsram,0x50);

  fseek(hand,0x474,SEEK_SET);
  fput(z80ram,   0x02000);

  fseek(hand,0x2478,SEEK_SET);
  byteswap_memory(ram,0x10000);
  fput(ram,      0x10000);
  byteswap_memory(ram,0x10000);

  fput(vdp.vram ,0x10000);

  return 0;
save_error:
  return 1;
}

#endif //_PSP10
