// DGen v1.13+
// Megadrive's VDP C++ module

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "md.h"

#ifdef _PSP10
static unsigned short new_highpal[ 64 ] ;
#endif //_PSP10

int md_vdp::get_screen_info(struct dgen_sinfo *si)
{
  // Release the pointers to vdp Data to an external function
  si->vram=vram; si->cram=cram; si->vsram=vsram;
  si->vdp_reg=reg;
  // NB - if you change anything, remember to set 'dirt' accordingly
  return 0;
}

int md_vdp::reset()
{
  if (!ok) return 1;

  rw_mode=0x00; rw_addr=0; rw_dma=0;
  memset(mem,0,0x10100);
  memset(reg,0,0x20);
  memset(dirt,0xff,0x35); // mark everything as changed
  return 0;
}

md_vdp::md_vdp()
{
}

void md_vdp::init()
{
  ok=0;
  belongs=0; // Don't know which megadrive vdp belongs to yet

  mem=vram=cram=vsram=NULL;
#ifdef _PSP10
  mem = md::md_mem_vdp ;
#else //_PSP10
  mem=new unsigned char[0x10100+0x35]; //0x20+0x10+0x4+1 for dirt
#endif //_PSP10
  if (mem==0) return;
  vram=mem+0x00000; cram=mem+0x10000; vsram=mem+0x10080;
  dirt=mem+0x10100; // VRAM/CRAM/Reg dirty buffer bitfield
  // Also in 0x34 are global dirt flags (inclduing VSRAM this time)

#ifdef _PSP10
  highpal = new_highpal ;
#else //_PSP10
  highpal=new unsigned int[64]; if (highpal==0) return;
#endif //_PSP10

  Bpp = Bpp_times8 = 0;

  ok=1;
  reset();

}

md_vdp::~md_vdp()
{
#ifndef _PSP10
  free(mem); 
#endif //_PSP10

  mem=vram=cram=vsram=NULL;
  ok=0;
}

void md_vdp::dma_clear()
{
	rw_dma  = 0 ;
//	rw_addr = 0 ;
}

int md_vdp::dma_len()
{ return (reg[0x14]<<8) | reg[0x13]; }

int md_vdp::dma_addr()
{
  int addr=0;
  addr =(reg[0x17]&0x7f)<<17;
  addr|=reg[0x16]<<9;
  addr|=reg[0x15]<<1;
  return addr;
}


// DMA can read from anywhere
unsigned char md_vdp::dma_mem_read(int addr)
{
  return belongs->misc_readbyte(addr);
}

inline unsigned long get_dword( unsigned char* src )
{
	unsigned long us ;
				
	us = *((unsigned long*)src) ;
	//us = *src << 8 | *(src+1) << 0 | *(src+2) << 24 | *(src+3) << 16  ;

	return ((us >> 8) & 0x00FF00FF) | ((us << 8) & 0xFF00FF00) ;
}

#define GET_WORD( src )  ((*((unsigned short*)src) << 8) | (*((unsigned short*)src) >> 8))
//#define GET_WORD( src )  (((*((unsigned short*)src) << 8)&0xFF00) | ((*((unsigned short*)src) >> 8)&0x00FF))


#if 0
void md_vdp::dma_copy( unsigned char* src, int len )
{
	/*
	int n, reg15 ;

	reg15 = reg[15] ;

	if( reg15 == 1 )
	{
		for( n=0; n < len ; n++ )
		{
			*rw_ptr = *src ;
			rw_ptr += 1 ;
			src    += 1 ;
		}

		dirt[0x34] |= rw_flag ;
	}
	else
	{
		for( n=0; n < len ; n++ )
		{
			*((unsigned short*)rw_ptr) = GET_WORD( src )  ;
			rw_ptr += reg15 ;
			src    += 2     ;
		}

		dirt[0x34] |= rw_flag ;
	}
	//*/

	//*
	int n, reg15 ;

	reg15 = reg[15] ;

	if( reg15 == 1 )
	{
//_st_mode :
		unsigned short us ;

		switch(rw_mode)
		{
		case 0x04: 
			for( n=0; n < len ; n++ )
			{
				if( vram[ rw_addr & 0xffff ] != *src )
				{
					vram[ rw_addr & 0xffff ] = *src ;
					dirt[0x34]|=1;

					rw_addr += 1 ;
					src     += 1 ;
					n++          ;
					for( ; n < len ; n++ )
					{
						vram[ rw_addr & 0xffff ] = *src ; //(*((unsigned short*)src) << 8) | (*((unsigned short*)src) >> 8) ;

						rw_addr += 1 ;
						src     += 1 ;
					}

					return ;
				}

				rw_addr += 1 ;
				src     += 1 ;
			}
			break ;
		case 0x0c: 
			for( n=0; n < len ; n++ )
			{
				if( cram[ rw_addr & 0x007f ] != *src )
				{
					cram[ rw_addr & 0x007f ] = *src ;
					dirt[0x34]|=2;

					rw_addr += 1 ;
					src     += 1 ;
					n++          ;
					for( ; n < len ; n++ )
					{
						cram[ rw_addr & 0x007f ] = *src ;

						rw_addr += 1 ;
						src     += 1 ;
					}

					return ;
				}

				rw_addr += 1 ;
				src     += 1 ;
			}
			break ;
		case 0x14: 
			for( n=0; n < len ; n++ )
			{
				if( vsram[ rw_addr & 0x007f ] != *src )
				{
					vsram[ rw_addr & 0x007f ] = *src ;
					dirt[0x34]|=4;

					rw_addr += 1 ;
					src     += 1 ;
					n++          ;
					for( ; n < len ; n++ )
					{
						vsram[ rw_addr & 0x007f ] = *src ; //(*((unsigned short*)src) << 8) | (*((unsigned short*)src) >> 8) ;

						rw_addr += 1 ;
						src     += 1 ;
					}

					return ;
				}

				rw_addr += 1 ;
				src     += 1 ;
			}
			break ;
		}
	}
	else //if( reg15 == 2 )
	{
		unsigned short us ;

		switch(rw_mode)
		{
		case 0x04: 
			for( n=0; n < len ; n++ )
			{
				us = GET_WORD( src ) ; 
				if( *((unsigned short*)&vram[ rw_addr ]) != us )
				{
					*((unsigned short*)&vram[ rw_addr ]) = us ;
					dirt[0x34]|=1;

					rw_addr += reg15 ;
					src     += 2     ;
					n++              ;
					for( ; n < len ; n++ )
					{
						*((unsigned short*)&vram[ rw_addr ]) = GET_WORD( src ) ; 

						rw_addr += reg15 ;
						src     += 2     ;
					}

					return ;
				}

				rw_addr += reg15 ;
				src     += 2     ;
			}
			break ;
		case 0x0c: 
			for( n=0; n < len ; n++ )
			{
				us = GET_WORD( src ) ; 
				if( *((unsigned short*)&cram[ rw_addr ]) != us )
				{
					*((unsigned short*)&cram[ rw_addr ]) = us ;
					dirt[0x34]|=2;

					rw_addr += reg15 ;
					src     += 2     ;
					n++              ;
					for( ; n < len ; n++ )
					{
						*((unsigned short*)&cram[ rw_addr ]) = GET_WORD( src ) ; 

						rw_addr += reg15 ;
						src     += 2     ;
					}

					return ;
				}

				rw_addr += reg15 ;
				src     += 2     ;
			}
			break ;
		case 0x14: 
			for( n=0; n < len ; n++ )
			{
				us = GET_WORD( src ) ; 
				if( *((unsigned short*)&vsram[ rw_addr ]) != us )
				{
					*((unsigned short*)&vsram[ rw_addr ]) = us ;
					dirt[0x34]|=4;

					rw_addr += reg15 ;
					src     += 2     ;
					n++              ;
					for( ; n < len ; n++ )
					{
						*((unsigned short*)&vsram[ rw_addr ]) = GET_WORD( src ) ; 

						rw_addr += reg15 ;
						src     += 2     ;
					}

					return ;
				}

				rw_addr += reg15 ;
				src     += 2     ;
			}
			break ;
		}
	}
	//*/
}

#else
void md_vdp::dma_copy( unsigned char* src, int len )
{
	/*
	int n, reg15 ;

	reg15 = reg[15] ;

	if( reg15 == 1 )
	{
		for( n=0; n < len ; n++ )
		{
			*rw_ptr = *src ;
			rw_ptr += 1 ;
			src    += 1 ;
		}

		dirt[0x34] |= rw_flag ;
	}
	else
	{
		for( n=0; n < len ; n++ )
		{
			*((unsigned short*)rw_ptr) = GET_WORD( src )  ;
			rw_ptr += reg15 ;
			src    += 2     ;
		}

		dirt[0x34] |= rw_flag ;
	}
	//*/

	//*
	int n, reg15 ;
	unsigned short us ;

	reg15 = reg[15] ;

	if( reg15 == 1 )
	{
		switch(rw_mode)
		{
		case 0x04: 
			for( n=0; n < len ; n++ )
			{
				vram[ rw_addr & 0xffff ] = *src ;
				rw_addr += 1 ;
				src     += 1 ;
			}
			dirt[0x34]|=1;
			break ;
		case 0x0c: 
			for( n=0; n < len ; n++ )
			{
				cram[ rw_addr & 0x007f ] = *src ;
				rw_addr += 1 ;
				src     += 1 ;
			}
			dirt[0x34]|=2;
			break ;
		case 0x14: 
			for( n=0; n < len ; n++ )
			{
				vsram[ rw_addr & 0x007f ] = *src ;
				rw_addr += 1 ;
				src     += 1 ;
			}
			dirt[0x34]|=4;
			break ;
		}
	}
	else //if( reg15 == 2 )
	{
		switch(rw_mode)
		{
		case 0x04: 
			for( n=0; n < len ; n++ )
			{
				us = GET_WORD( src ) ; 
				*((unsigned short*)&vram[ rw_addr ]) = us ;
				rw_addr += reg15 ;
				src     += 2     ;
			}
			dirt[0x34]|=1;
			break ;
		case 0x0c: 
			for( n=0; n < len ; n++ )
			{
				us = GET_WORD( src ) ; 
				*((unsigned short*)&cram[ rw_addr ]) = us ;
				rw_addr += reg15 ;
				src     += 2     ;
			}
			dirt[0x34]|=2;
			break ;
		case 0x14: 
			for( n=0; n < len ; n++ )
			{
				us = GET_WORD( src ) ; 
				*((unsigned short*)&vsram[ rw_addr ]) = us ;
				rw_addr += reg15 ;
				src     += 2     ;
			}
			dirt[0x34]|=4;
			break ;
		}
	}
	//*/


}

#endif 

void md_vdp::dma_memset( unsigned char* src, unsigned d, int len )
{

}

// Must go through these calls to update the dirty flags
int md_vdp::poke_vram(int addr,unsigned char d)
{
// Keeping GCC happy over unused vars. [PKH]
//  int diff=0;
  addr&=0xffff;
  if (vram[addr]!=d)
  {
    // Store dirty information down to 256 byte level in bits
#ifdef _USE_DIRT
    int byt,bit;
    byt=addr>>8; bit=byt&7; byt>>=3; byt&=0x1f;
    dirt[0x00+byt]|=(1<<bit); 
#endif // _USE_DIRT
	dirt[0x34]|=1;
    vram[addr]=d;
  }
  return 0;
}
int md_vdp::poke_cram(int addr,unsigned char d)
{
//  int diff=0;
  addr&=0x007f;
  if (cram[addr]!=d)
  {
    // Store dirty information down to 1byte level in bits
#ifdef _USE_DIRT
    int byt,bit;
    byt=addr; bit=byt&7; byt>>=3; byt&=0x0f;
    dirt[0x20+byt]|=(1<<bit); 
#endif // _USE_DIRT
	dirt[0x34]|=2;
    cram[addr]=d; 
  }

  return 0;
}
int md_vdp::poke_vsram(int addr,unsigned char d)
{
//  int diff=0; 
  addr&=0x007f;
  if (vsram[addr]!=d)
  { dirt[0x34]|=4; vsram[addr]=d; }
  return 0;
}


// Must go through these calls to update the dirty flags
int md_vdp::poke_vram_s(int addr,unsigned short d)
{
// Keeping GCC happy over unused vars. [PKH]
//  int diff=0;
//  addr&=0xffff;
  d = (d>>8)|(d<<8) ;
  if (*((unsigned short*)&vram[addr])!=d)
  {
    // Store dirty information down to 256 byte level in bits
#ifdef _USE_DIRT
    int byt,bit;
    byt=addr>>8; bit=byt&7; byt>>=3; byt&=0x1f;
    dirt[0x00+byt]|=(3<<bit); 
#endif // _USE_DIRT
	dirt[0x34]|=1;
    *((unsigned short*)&vram[addr])=d;
  }
  return 0;
}
int md_vdp::poke_cram_s(int addr,unsigned short d)
{
//  int diff=0;
//  addr&=0x007f;
  d = (d>>8)|(d<<8) ;
  if (*((unsigned short*)&cram[addr])!=d)
  {
    // Store dirty information down to 1byte level in bits
#ifdef _USE_DIRT
    int byt,bit;
    byt=addr; bit=byt&7; byt>>=3; byt&=0x0f;
    dirt[0x20+byt]|=(3<<bit); 
#endif //_USE_DIRT
	dirt[0x34]|=2;
    *((unsigned short*)&cram[addr])=d; 
  }

  return 0;
}
int md_vdp::poke_vsram_s(int addr,unsigned short d)
{
//  int diff=0; 
//  addr&=0x007f;
  d = (d>>8)|(d<<8) ;
  if (*((unsigned short*)&vsram[addr])!=d)
  { dirt[0x34]|=4; *((unsigned short*)&vsram[addr])=d; }
  return 0;
}

int md_vdp::putword(unsigned short d)
{
	//*
//  int diff=0;
  // Called by dma or a straight write
#ifdef   _PSP10
  if( (rw_addr % 2) == 0 )
  {
	  switch(rw_mode)
	  {
		case 0x04: poke_vram_s (rw_addr&0xFFFF,d); break;//poke_vram (rw_addr+0,d>>8); poke_vram (rw_addr+1,d&0xff); break;
		case 0x0c: poke_cram_s (rw_addr&0x007F,d); break;//poke_cram (rw_addr+0,d>>8); poke_cram (rw_addr+1,d&0xff); break;
		case 0x14: poke_vsram_s(rw_addr&0x007F,d); break;//poke_vsram(rw_addr+0,d>>8); poke_vsram(rw_addr+1,d&0xff); break;
	  }
  }
  else
//#else //_PSP10
#endif //_PSP10
  {
	  switch(rw_mode)
	  {
		case 0x04: poke_vram ((rw_addr+0)&0xFFFF,d>>8); poke_vram ((rw_addr+1)&0xFFFF,d&0xff); break;
		case 0x0c: poke_cram ((rw_addr+0)&0x007F,d>>8); poke_cram ((rw_addr+1)&0x007F,d&0xff); break;
		case 0x14: poke_vsram((rw_addr+0)&0x007F,d>>8); poke_vsram((rw_addr+1)&0x007F,d&0xff); break;
	  }
  }
//#endif //_PSP10

  rw_addr+=reg[15];
  return 0;
  //*/

  /*
	d = (d>>8)|(d<<8) ;
	if( *((unsigned short*)rw_ptr) != d )
	{
		*((unsigned short*)rw_ptr) = d ;
		dirt[0x34] |= rw_flag ;
	}
	rw_ptr += reg[15] ;
	return 0 ;
	*/
}

int md_vdp::putbyte(unsigned char d)
{
	//*
//  int diff=0;
  // Called by dma or a straight write
  switch(rw_mode)
  {
    case 0x04: poke_vram (rw_addr,d>>8); break;
    case 0x0c: poke_cram (rw_addr,d>>8); break;
    case 0x14: poke_vsram(rw_addr,d>>8); break;
  }
  rw_addr+=reg[15];
  return 0;
  //*/

	/*
	if( *rw_ptr != d )
	{
		*rw_ptr = d ;
		dirt[0x34] |= rw_flag ;
	}
	rw_ptr += reg[15] ;
	return 0 ;
	//*/
}

#undef MAYCHANGE

unsigned short md_vdp::readword()
{
	//*
  // Called by a straight read only
  unsigned short result;
#ifdef _PSP10

  if( (rw_addr % 2) == 0 )
  {
	  switch(rw_mode)
	  {
		case 0x00: result=*((unsigned short*)&vram [rw_addr&0xffff]); break;
		case 0x20: result=*((unsigned short*)&cram [rw_addr&0x007f]); break;
		case 0x10: result=*((unsigned short*)&vsram[rw_addr&0x007f]); break;
	  }

	  result = (result << 8) | (result >> 8) ;
  }
  else
  {
	  switch(rw_mode)
	  {
		case 0x00: result=( vram[(rw_addr+0)&0xffff]<<8)+ vram[(rw_addr+1)&0xffff]; break;
		case 0x20: result=( cram[(rw_addr+0)&0x007f]<<8)+ cram[(rw_addr+1)&0x007f]; break;
		case 0x10: result=(vsram[(rw_addr+0)&0x007f]<<8)+vsram[(rw_addr+1)&0x007f]; break;
	  }
  }
  rw_addr+=reg[15];
  return result ;
#else //_PSP10
  switch(rw_mode)
  {
    case 0x00: result=( vram[(rw_addr+0)&0xffff]<<8)+
                        vram[(rw_addr+1)&0xffff]; break;
    case 0x20: result=( cram[(rw_addr+0)&0x007f]<<8)+
                        cram[(rw_addr+1)&0x007f]; break;
    case 0x10: result=(vsram[(rw_addr+0)&0x007f]<<8)+
                       vsram[(rw_addr+1)&0x007f]; break;
  }
  rw_addr+=reg[15];
  return result;
#endif //_PSP10
  //*/

//	unsigned short ret = *((unsigned short*)rw_ptr) ;
//	rw_ptr += reg[15] ;
//	return (ret << 8) | (ret >> 8) ;
}

unsigned char md_vdp::readbyte()
{
  // Called by a straight read only
	//*
  unsigned char result;
  switch(rw_mode)
  {
#ifdef _PSP10
    case 0x00: result= vram[(rw_addr)&0xffff]; break;
    case 0x20: result= cram[(rw_addr)&0x007f]; break;
    case 0x10: result=vsram[(rw_addr)&0x007f]; break;
#else //_PSP10
    case 0x00: result= vram[(rw_addr+0)&0xffff]; break;
    case 0x20: result= cram[(rw_addr+0)&0x007f]; break;
    case 0x10: result=vsram[(rw_addr+0)&0x007f]; break;
#endif //_PSP10
  }
  rw_addr+=reg[15];
  return result;
  //*/

	//unsigned char ret = *rw_ptr ;
	//rw_ptr += reg[15] ;
	//return ret ;
}

int md_vdp::command(unsigned int cmd)
{
	// Decode 32-bit VDP command
	rw_dma=((cmd&0x80)==0x80);
	rw_mode= cmd&0x00000070;
	rw_mode|=(cmd&0xc0000000)>>28;
	// mode writes: 04=VRAM 0C=CRAM 14=VSRAM
	// mode reads:  00=VRAM 20=CRAM 10=VSRAM
	rw_addr= (cmd&0x00000003)<<14;
	rw_addr|=(cmd&0x3fff0000)>>16;
	// If not dma (or we need a fill),
	// we are set up to write any data sent to vdp data reg

	// if it's a dma request do it straight away
	if (rw_dma)
	//if( reg[1] & 0x10 )
	{
		int i   ;
		int s   = dma_addr() ; 
		int len = dma_len () ;

		switch ( (reg[0x17]>>6)&3 )
		{
		case 0: case 1:
			if( s > 0x00dfffff )
			{
				dma_copy( &md::ram[s&0xffff], len ) ;
			}
			else if( s < MAX_ROM_SIZE )
			{
				dma_copy( &md::rom[s], len ) ;
			}
			else
			{
				switch(rw_mode)
				{
				case 0x04: 
					for (i=0;i<len;i++)
					{
						poke_vram_s (rw_addr, belongs->misc_readword(s) ); 
						rw_addr+=reg[15];
						s += 2;
					}
					break;
				case 0x0c: 
					for (i=0;i<len;i++)
					{
						poke_cram_s (rw_addr, belongs->misc_readword(s) ); 
						rw_addr+=reg[15];
						s += 2;
					}
					break;
				case 0x14: 
					for (i=0;i<len;i++)
					{
						poke_vsram_s (rw_addr, belongs->misc_readword(s) ); 
						rw_addr+=reg[15];
						s += 2;
					}
					break;
				}
			}
			break;
		case 2:
			// Done later on
			break;
		case 3:
			//*
			for( i=0; i < len ; i++ )
			{
				unsigned short val;
				val= vram[(s++)&0xffff]; val<<=8;
				val|=vram[(s++)&0xffff]; 
				putword(val);
			}
			//*/
			/*
			for( i=0; i < len ; i++ )
			{
				vram[ rw_addr & 0xFFFF] = vram[ s & 0xFFFF ] ;
				rw_addr+=reg[15];
				s++;
			}
			//*/
			break;
		}
	}

	return 0;
}

int md_vdp::writeword(unsigned short d)
{
  putword(d);
  //poke_vram ( rw_addr,d&0xFF   ); 
  //rw_addr+=reg[15];

  if (rw_dma)
  //if( reg[1] & 0x10 )
  {
	  //rw_dma = 0 ;
    // This is the 'done later on' bit for words
    // Do a dma fill if it's set up:
    if (((reg[0x17]>>6)&3)==2)
    {
      int i,len ;
      len=dma_len();
#ifdef _PSP10

	  //*
			d = (d >> 8) & 0xFF ;
			for (i=0;i<len;i++)
			{
				poke_vram( rw_addr^1, d ); 
				rw_addr+=reg[15];
			}
	  //*/

	  /*
	  if( (reg[15] == 1) || (rw_addr % 2) )
	  { 
		switch(rw_mode)
		{
		case 0x04: 
			for (i=0;i<len;i++)
			{
				poke_vram ((rw_addr  )&0xFFFF,d>>8   ); 
				poke_vram ((rw_addr+1)&0xFFFF,d&0xFF ); 
				rw_addr+=reg[15];
			}
			break;
		case 0x0c: 
			for (i=0;i<len;i++)
			{
				poke_cram ( (rw_addr  )&0x7F,d>>8  ); 
				poke_cram ( (rw_addr+1)&0x7F,d&0xFF); 
				rw_addr+=reg[15];
			}
			break;
		case 0x14: 
			for (i=0;i<len;i++)
			{
				poke_vsram ((rw_addr  ),d>>8  ); 
				poke_vsram ((rw_addr+1),d&0xFF); 
				rw_addr+=reg[15];
			}
			break;
		default:
			for (i=0;i<len;i++){ rw_addr+=reg[15]; }
			break ;
		}
	  }
	  else if( reg[15] == 2 )
	  {
		switch(rw_mode)
		{
		case 0x04: 
			for (i=0;i<len;i++)
			{
				poke_vram_s (rw_addr,d ); 
				rw_addr+=reg[15];
			}
			break;
		case 0x0c: 
			for (i=0;i<len;i++)
			{
				poke_cram_s (rw_addr,d ); 
				rw_addr+=reg[15];
			}
			break;
		case 0x14: 
			for (i=0;i<len;i++)
			{
				poke_vsram_s (rw_addr,d ); 
				rw_addr+=reg[15];
			}
			break;
		default:
			for (i=0;i<len;i++){ rw_addr+=reg[15]; }
			break ;
		}
	  }
	  else
	  {
		  for (i=0;i<len;i++)
			putword(d);
	  }
	  //*/

#else //_PSP10
      for (i=0;i<len;i++)
        putword(d);
        //putbyte(d);
#endif //_PSP10

	  //reg[1] &= ~0x10 ;
      return 0;
    }
  }
  else
  {
//    putword(d);
    return 0;
  }
  return 0;
}

int md_vdp::writebyte(unsigned char d)
{
  putbyte(d);

  if (rw_dma)
//  if( reg[1] & 0x10 )
  {
	//  rw_dma = 0 ;
    // This is the 'done later on' bit for bytes
    // Do a dma fill if it's set up:
    if (((reg[0x17]>>6)&3)==2)
    {
	  //*
      int i,len;
      len=dma_len();

		d = d >> 8 ;
		for (i=0;i<len;i++)
		{
			poke_vram( rw_addr, d ); 
			rw_addr+=reg[15];
		}
	  //*/

		/*
      int i,len;
      len=dma_len();
      for (i=0;i<len;i++)
        putbyte(d);
		//*/
      return 0;
    }
  }
  else
  {
    //putbyte(d);
    return 0;
  }

  return 0;
}

