///////////////////////////////////////////////////////////////////////////
/// psp_main.c
///////////////////////////////////////////////////////////////////////////
#include "emu_core.h"
#include "psp_main.h"
#include "psp_screen.h"
#include "psp_menu.h"
#include "psp_pad.h"
#include "psp_ex.h"

#include <pspkernel.h>
#include <pspdebug.h>
#include <psppower.h>

PSP_MODULE_INFO("dgen_psp", 0, 1, 1);
PSP_MAIN_THREAD_ATTR(THREAD_ATTR_USER);

/* Exit callback */
int exit_callback(int arg1, int arg2, void *common)
{
	sceKernelExitGame();

	return 0;
}

/* Callback thread */
int CallbackThread(SceSize args, void *argp)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);

	sceKernelSleepThreadCB();

	return 0;
}

/* Sets up the callback thread and returns its thread id */
int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}

///////////////////////////////////////////////////////////////////////////
int           PSP_Power_Init           ( void ) ;
int           PSP_Power_CheckSleep     ( void ) ;
int           PSP_Sound_Init           ( int nChannel ) ;
int           PSP_Ctrlpad_Init         ( void ) ;
PSP_PAD_DATA* PSP_Ctrlpad_GetPadData   () ;
void          PSP_Sound_Enable         ( short nCh, short nEnable );
///////////////////////////////////////////////////////////////////////////
#define MAX_PSP_SOUND_BUFFER   (PSP_SOUND_SAMPLE * (2+2) * 2)
int psp_sound_vol = (0) ;
///////////////////////////////////////////////////////////////////////////
EMU_CONFIG* emc ;
///////////////////////////////////////////////////////////////////////////
#define _OVER_CLOCK 
//#define _DEBUG 


#ifdef _DEBUG
unsigned long tm_debug[16] ;
#endif //_DEBUG

int debug1, debug2 ;

unsigned long key_menu  = CTRL_LTRIGGER ;
unsigned long tm_ed     = 0         ;
unsigned long tm_pd     = 0         ;
unsigned long tm_60     = 0         ;
unsigned long tm_st     = 0         ;
unsigned long tm_run    = TM_1FRAME ;
unsigned long tm_skip   = TM_1FRAME ;
         long tm_res    = 0         ;
unsigned char skip      = 0         ;
unsigned char skip_s    = 0         ;
unsigned char skip_t    = 0         ;
unsigned char fps       = 0         ;
unsigned char fps_last  = 0         ;
signed   char exec_cfg  = 0         ;
char          exec_cfg_msg[32]      ;

unsigned long menu_keys[] =
{
	CTRL_LTRIGGER               ,
	CTRL_LTRIGGER | CTRL_START  ,
	CTRL_LTRIGGER | CTRL_SELECT ,
	CTRL_RTRIGGER               ,
	CTRL_RTRIGGER | CTRL_START  ,
	CTRL_RTRIGGER | CTRL_SELECT ,
	CTRL_START                  ,
	CTRL_SELECT                  
} ;

inline void __wait( unsigned long tm )
{
	unsigned long t ;

	t = sceKernelLibcClock() ;
	while( (sceKernelLibcClock() - t) < tm ){ ; }
}

/*
unsigned char calc_skip()
{
	static long t_rest ;
	int n, r ;
	long t, tt, min_tt, min_n, min_t ;
	int max = emc->max_fps / 2 ;

	min_tt = (TM_1FRAME * 60 / 2) ;
	min_n  = 0 ;
	min_t  = 0 ;

#ifdef _DEBUG
	if( (tm_debug[8] > tm_run) || (tm_debug[8] == 0)  ){ tm_debug[ 8] = tm_run  ; }
	else if( tm_debug[ 9] < tm_run  ){ tm_debug[ 9] = tm_run  ; }
	if( (tm_debug[10] > tm_skip) || (tm_debug[10] == 0) ){ tm_debug[10] = tm_skip ; }
	else if( tm_debug[11] < tm_skip ){ tm_debug[11] = tm_skip ; }
#endif // _DEBUG

	for( n=0; n <= emc->skip_count ; n++ )
	{
		r = max / (n+1) ;
		t = r * tm_run + (max - r) * tm_skip ;
		t = ((TM_1FRAME * 60 / 2) - t_rest) - t ;

		if( t < 0 ){ tt = -t ; }else{ tt = t ; }

		if( tt < min_tt ){ min_tt = tt ; min_n = n ; min_t = t ; }
	}

	t_rest = min_t ;

#ifdef _DEBUG
	tm_debug[3] = t_rest ;
	tm_debug[4] = min_t ;
	tm_debug[5] = min_n ;
#endif //_DEBUG

	if(      min_n < skip_s ){ skip = 0 ; }
	else if( min_n > skip_s ){ if(min_n > 1){ skip = 1 ; }else{ skip = 0 ;} }

	return min_n ;
}
*/

unsigned char calc_skip( unsigned long tm_frame )
{
	tm_res += ((TM_1FRAME*60) / emc->max_fps) - (long)tm_frame ;

	if( tm_res < 0 )
	{
		if( skip_t < skip_s ){ skip_t++ ; return 1; }
		skip_t = 0 ;
		tm_res = 0 ;
	}
	else if( tm_res > TM_1FRAME*2 )
	{
		tm_res = TM_1FRAME*2 ;
	}

	return 0 ;
}

inline void resize_image_320to80( unsigned short* src, unsigned short* dst )
{
	int x, y ;

	for( y=0; y < 56 ; y++ )
	{
		for( x=0; x < 80 ; x++ )
		{
			dst[x] = src[x*4] ;
		}

		dst += 80 ;
		src += PSP_LINESIZE * 4 ;
	}
}

inline int EmuCfg_ExecKey( int pad_cfg )
{
	switch( pad_cfg )
	{
	case MD_PAD_SKIP_U    : 
		emc->skip_count++  ; 
		if(emc->skip_count > MAX_FRAME_SKIP){ emc->skip_count = MAX_FRAME_SKIP ; } 
		Menu_GetMsg_Skip( exec_cfg_msg ) ;
		break ;
	case MD_PAD_SKIP_D    : 
		emc->skip_count--  ; 
		if(emc->skip_count < MIN_FRAME_SKIP){ emc->skip_count = MIN_FRAME_SKIP ; } 
		Menu_GetMsg_Skip( exec_cfg_msg ) ;
		break ;
	case MD_PAD_SCREEN    : 
		emc->screen_mode++ ; 
		if(emc->screen_mode > SCREEN_FULL){ emc->screen_mode = SCREEN_1X ; } 
		Menu_DrawBG() ;
		Menu_GetMsg_Screen( exec_cfg_msg ) ;
		break ;
	case MD_PAD_FPS       : 
		emc->show_fps++ ; 
		if( emc->show_fps > SHOW_FPS_RB ){ emc->show_fps = 0 ; } 
		Menu_DrawBG() ;
		Menu_GetMsg_FPS( exec_cfg_msg ) ;
		break ;
	case MD_PAD_VSYNC     : 
		emc->show_flags ^= SHOW_VSYNC ;
		Menu_GetMsg_Vsync( exec_cfg_msg ) ;
		break ;
	case MD_PAD_SOUND_U   : 
		emc->sound_on++ ; 
		if( emc->sound_on > MD_SOUND_44K ){ emc->sound_on = MD_SOUND_44K ; }
		Menu_GetMsg_SoundMode( exec_cfg_msg ) ;
		break ;
	case MD_PAD_SOUND_D   : 
		emc->sound_on-- ; 
		if( emc->sound_on < MD_SOUND_OFF ){ emc->sound_on = MD_SOUND_OFF ; }
		Menu_GetMsg_SoundMode( exec_cfg_msg ) ;
		break ;
	case MD_PAD_SOUND_V_U : 
		emc->sound_vol++ ;
		if( emc->sound_vol > VOL_16 ){ emc->sound_vol = VOL_16 ; }
		Menu_GetMsg_SoundVol( exec_cfg_msg ) ;
		break ;
	case MD_PAD_SOUND_V_D : 
		emc->sound_vol-- ;
		if( emc->sound_vol < VOL_08 ){ emc->sound_vol = VOL_08 ; }
		Menu_GetMsg_SoundVol( exec_cfg_msg ) ;
		break ;
	case MD_PAD_STATE_SUB: // select state slot 
		emu_state.cfg.slot_sel-- ;
		if( emu_state.cfg.slot_sel < 0 ){ emu_state.cfg.slot_sel = MAX_STATE_SLOT-1; }
		Menu_GetMsg_State( exec_cfg_msg, 0, emu_state.cfg.slot_sel ) ;
		break ; 
	case MD_PAD_STATE_ADD: // select state slot 
		emu_state.cfg.slot_sel++ ;
		if( emu_state.cfg.slot_sel > (MAX_STATE_SLOT-1) ){ emu_state.cfg.slot_sel = 0 ; }
		Menu_GetMsg_State( exec_cfg_msg, 0, emu_state.cfg.slot_sel ) ;
		break ; 
	case MD_PAD_STATE_LD : // load state
		if( EmuStateSlot_Load_MEM( &emu_state.slot[ emu_state.cfg.slot_sel ] ) )
		{
			Menu_GetMsg_State( exec_cfg_msg, 1, emu_state.cfg.slot_sel ) ;
		}
		else{ return 0 ; }
		break ; 
	case MD_PAD_STATE_SV : // save state
		resize_image_320to80( 
			emc->screen_mode == SCREEN_1X ? PSP_Screen_GetBackBuffer( 80, 16+8 ) : PSP_Screen_GetTempBuffer( 16, 16 ), 
			&state_temp_image[0][0] ) ;

		EmuState_SaveSlot( &emu_state, emu_state.cfg.slot_sel, &state_temp_image[0][0], path_state ) ;
		Menu_GetMsg_State( exec_cfg_msg, 2, emu_state.cfg.slot_sel ) ;
		break ; 
	case MD_PAD_STATE_LDL: // load last slot
		if( EmuStateSlot_Load_MEM( &emu_state.slot[ emu_state.cfg.slot_inc ] ) )
		{
			Menu_GetMsg_State( exec_cfg_msg, 3, emu_state.cfg.slot_inc ) ;
		}
		else{ return 0 ; }
		break ; 
	case MD_PAD_STATE_SVL: // save slot+
		resize_image_320to80( 
			emc->screen_mode == SCREEN_1X ? PSP_Screen_GetBackBuffer( 80, 16+8 ) : PSP_Screen_GetTempBuffer( 16, 16 ), 
			&state_temp_image[0][0] ) ;

		EmuState_CalcInc( &emu_state ) ; 
		EmuState_SaveSlot( &emu_state, emu_state.cfg.slot_inc, &state_temp_image[0][0], path_state ) ;
		Menu_GetMsg_State( exec_cfg_msg, 4, emu_state.cfg.slot_inc ) ;
		break ; 
	case MD_PAD_CHEAT_ON :
		emu_cheat.all_on = emu_cheat.all_on ? 0 : 1 ;
		Menu_GetMsg_CheatOn( exec_cfg_msg, emu_cheat.all_on ) ;
		break ;
	default : return 0 ;
	}

	return 1 ;
}

inline int exec_key_cfg( PSP_PAD_DATA* pPad )
{
	if( pPad->buttons & CTRL_LTRIGGER )
	{
		if( emc->pad_cfg[MD_PAD_CFG_L] )
		{
			if( EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_L] ) ){ return 1 ; }
		}
	}

	if( pPad->buttons & CTRL_RTRIGGER )
	{
		if( emc->pad_cfg[MD_PAD_CFG_R] )
		{
			if( EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R] ) ){ return 1 ; }
		}
		else
		{
			if( (pPad->buttons & CTRL_CIRCLE  ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_CIRCLE  ] ) ){ return 1 ; }
			if( (pPad->buttons & CTRL_CROSS   ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_CROSS   ] ) ){ return 1 ; }
			if( (pPad->buttons & CTRL_TRIANGLE) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_TRIANGLE] ) ){ return 1 ; }
			if( (pPad->buttons & CTRL_SQUARE  ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_SQUARE  ] ) ){ return 1 ; }
			if( (pPad->buttons & CTRL_SELECT  ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_SELECT  ] ) ){ return 1 ; }
			if( (pPad->buttons & CTRL_START   ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_START   ] ) ){ return 1 ; }

			if( emc->analog_mode == PSP_ANALOG_SWAP )
			{
				if( (pPad->buttons & CTRL_UP   ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_DPAD_T] ) ){ return 1 ; }
				if( (pPad->buttons & CTRL_DOWN ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_DPAD_B] ) ){ return 1 ; }
				if( (pPad->buttons & CTRL_LEFT ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_DPAD_L] ) ){ return 1 ; }
				if( (pPad->buttons & CTRL_RIGHT) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_DPAD_R] ) ){ return 1 ; }
			}
			else
			{
				if( (pPad->analog[CTRL_ANALOG_Y] < LOWER_THRESHOLD) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_DPAD_T] ) ){ return 1 ; }
				if( (pPad->analog[CTRL_ANALOG_Y] > UPPER_THRESHOLD) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_DPAD_B] ) ){ return 1 ; }
				if( (pPad->analog[CTRL_ANALOG_X] < LOWER_THRESHOLD) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_DPAD_L] ) ){ return 1 ; }
				if( (pPad->analog[CTRL_ANALOG_X] > UPPER_THRESHOLD) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_R_DPAD_R] ) ){ return 1 ; }
			}
		}
	}
	else
	{
		if( (pPad->buttons & CTRL_CIRCLE  ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_CIRCLE  ] ) ){ return 1 ; }
		if( (pPad->buttons & CTRL_CROSS   ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_CROSS   ] ) ){ return 1 ; }
		if( (pPad->buttons & CTRL_TRIANGLE) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_TRIANGLE] ) ){ return 1 ; }
		if( (pPad->buttons & CTRL_SQUARE  ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_SQUARE  ] ) ){ return 1 ; }
		if( (pPad->buttons & CTRL_SELECT  ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_SELECT  ] ) ){ return 1 ; }
		if( (pPad->buttons & CTRL_START   ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_START   ] ) ){ return 1 ; }

		if( emc->analog_mode == PSP_ANALOG_SWAP )
		{
			if( (pPad->buttons & CTRL_UP   ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_DPAD_T] ) ){ return 1 ; }
			if( (pPad->buttons & CTRL_DOWN ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_DPAD_B] ) ){ return 1 ; }
			if( (pPad->buttons & CTRL_LEFT ) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_DPAD_L] ) ){ return 1 ; }
			if( (pPad->buttons & CTRL_RIGHT) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_DPAD_R] ) ){ return 1 ; }
		}
		else if( emc->analog_mode == PSP_ANALOG_CFG )
		{
			if( (pPad->analog[CTRL_ANALOG_Y] < LOWER_THRESHOLD) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_DPAD_T] ) ){ return 1 ; }
			if( (pPad->analog[CTRL_ANALOG_Y] > UPPER_THRESHOLD) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_DPAD_B] ) ){ return 1 ; }
			if( (pPad->analog[CTRL_ANALOG_X] < LOWER_THRESHOLD) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_DPAD_L] ) ){ return 1 ; }
			if( (pPad->analog[CTRL_ANALOG_X] > UPPER_THRESHOLD) && EmuCfg_ExecKey( emc->pad_cfg[MD_PAD_CFG_DPAD_R] ) ){ return 1 ; }
		}
	}

	return 0 ;
}

inline void ShowFPS( signed char pos, unsigned char fps )
{
//	const int show_x[] = { 1, 1, PSP_SCREEN_WIDTH - 5*6 -1, 1, PSP_SCREEN_WIDTH - 5*6 -1 } ;
//	const int show_y[] = { 1, 1, 1, PSP_SCREEN_HEIGHT -11, PSP_SCREEN_HEIGHT -11 } ;
	const int show_x[] = { 80  , 80  , 80 + 320 - 5*6 -1, 80, 80+ 320 - 5*6 -1 } ;
	const int show_y[] = { 16-2, 16-2, 16-2, 16-7+240 , 16-7+240 } ;

	PSP_Screen_DrawInt2( show_x[pos]     , show_y[pos] , fps    , 0xFFFF ) ;
	PSP_Screen_DrawText( show_x[pos] +5*2, show_y[pos] , "FPS " , 0xFFFF ) ;
}

inline void ShowFps2( signed char pos, unsigned char fps )
{
	const int show_x[] = { 0, 0, 480 - 5*5, 0, 480 - 5*5 } ;
	const int show_y[] = { 0, 0, 0, 272 -9, 272 -9 } ;

	PSP_Screen_DrawInt2( show_x[pos]      , show_y[pos] , fps    , 0xFFFF ) ;
	PSP_Screen_DrawText( show_x[pos] + 5*2, show_y[pos] , "FPS " , 0xFFFF ) ;
}

inline void ShowExecCfgMsg()
{
//	PSP_Screen_FillRect( 80,     16-2, PSP_SCREEN_WIDTH - 80,     16-2 +9, 0 ) ;
	PSP_Screen_FillRect( 80, 16-7+240, PSP_SCREEN_WIDTH - 80, 16-7+240 +9, 0 ) ;
	if( exec_cfg > 20 ){ PSP_Screen_DrawText( 240 - strlen(exec_cfg_msg) * 5 / 2, 16-7+240, exec_cfg_msg, 0xFFFF ) ; }
}

inline void ShowExecCfgMsg2()
{
	PSP_Screen_DrawText( 240 - strlen(exec_cfg_msg) * 5 / 2, 272-9, exec_cfg_msg, 0xFFFF ) ;
}

inline int UpdateTimer()
{
	int ret= 0 ;
	fps++ ;
	tm_st = tm_ed ;

	if( (tm_ed - tm_pd) > TM_1FRAME )
	{
		tm_pd = tm_ed ;
		PSP_Ctrlpad_Update() ;

		if( (PSP_Ctrlpad_GetButtons() & key_menu) == key_menu )
		{ 
			resize_image_320to80( PSP_Screen_GetBackBuffer( 80, 16+8 ), &state_temp_image[0][0] ) ;
			ret = 2 ;
		}
		else
		{
			if( exec_cfg > 0 )
			{
				exec_cfg-- ;
				md_setpaddata( PSP_Ctrlpad_GetPadData(), emc ) ; 
			}
			else if( exec_key_cfg( PSP_Ctrlpad_GetPadData() ) )
			{
				exec_cfg = 60 ;
				ret = 1 ;
			}
			else
			{
				md_setpaddata( PSP_Ctrlpad_GetPadData(), emc ) ; 
			}
		}
	}
	if( (tm_ed - tm_60) > (TM_1FRAME*60) )
	{
		fps_last = fps   ;
		fps      = 0     ;
		tm_60    = tm_ed ;

#ifdef _DEBUG
		PSP_Screen_DrawIntN( 1, 10, tm_run      , 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1, 20, tm_skip     , 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1, 30, tm_debug[ 2], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1, 60, tm_debug[ 3], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1, 70, tm_debug[ 4], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1, 80, tm_debug[ 5], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1, 90, tm_debug[ 6], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1,100, tm_debug[ 7], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1,110, tm_debug[ 8], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1,120, tm_debug[ 9], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1,130, tm_debug[10], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1,140, tm_debug[11], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1,150, tm_debug[12], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1,160, tm_debug[13], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1,170, tm_debug[14], 8, 0xFFFF ) ;
		PSP_Screen_DrawIntN( 1,180, tm_debug[15], 8, 0xFFFF ) ;

		tm_debug[4] = tm_debug[5] = tm_debug[6] = tm_debug[7] = 0 ;
		tm_debug[8] = tm_debug[9] = tm_debug[10] = tm_debug[11] = 0 ;
#endif //_DEBUG

	}

	//skip_s = calc_skip() ;

	EmuCheat_Exec( &emu_cheat ) ;

	return ret ;
}

int main(int argc, char *argv)
{
	EmuCfg_init( &emu_config[ EMU_MD  ] ) ;
	EmuCfg_init( &emu_config[ EMU_PCE ] ) ;

	Menu_Init        ( argc, "ms0:/PSP/GAME/DGEN/") ;
	//PSP_Power_Init   ( ) ;
	SetupCallbacks();
	PSP_Screen_Init  ( ) ;
	PSP_Sound_Init   ( 1 ) ;
	PSP_Ctrlpad_Init ( ) ;

#ifdef _USE_MD
	md_init_memory ( sEmuCoreBuff, MAX_EMU_CORE_BUFF ) ;
	md_init  () ;
#endif //_USE_MD

#ifdef _USE_PCE
	pce_init_memory( sEmuCoreBuff, MAX_EMU_CORE_BUFF ) ;
	pce_init () ;
#endif //_USE_PCE

_MAIN_MENU :
	PSP_Sound_Enable(0,0) ;
	scePowerSetClockFrequency(222,222,111);
	switch( Menu_MainMenu () )
	{
	default:
	case MENU_CMD_CANCEL   : 
		break ;

#ifdef _USE_MD
	case MENU_CMD_LOAD_MD       : 
	case MENU_CMD_LOAD_MD_ROM_S :
		if( md_load_rom( path_rom ) == 0 ){ emu_mode = EMU_MD ; }
		else { goto _MAIN_MENU ; }

		md_load_sram( path_sram ) ;

		emc = &emu_config[ emu_mode  ] ;
		EmuCfg_Load( emc, path_cfg ) ;

		EmuCheat_Load( &emu_cheat, path_cheat ) ;

		if( menu_cmd == MENU_CMD_LOAD_MD_ROM_S )
		{
			//md_run_skip() ;
			//EmuStateSlot_Load_MEM( &emu_state.slot[ menu_load_state_slot ] ) ;
			for( fps=0; fps < 120; fps++ ){ md_run_skip() ; }
			EmuStateSlot_Load_MEM( &emu_state.slot[ menu_load_state_slot ] ) ;
		}
		Menu_DrawBG() ;
		break ;
	case MENU_CMD_LOAD_MD_STATE :
		EmuStateSlot_Load_MEM( &emu_state.slot[ menu_load_state_slot ] ) ;
		break ;
#endif //_USE_MD

#ifdef _USE_PCE
	case MENU_CMD_LOAD_PCE : 
		if( pce_load_rom( path_rom ) == 0 )
		{
			emu_mode = EMU_PCE ;
		}
		else { goto _MAIN_MENU ; }
		emc = &emu_config[ emu_mode  ] ;
		EmuCfg_Load( emc, path_cfg ) ;
		Menu_DrawBG() ;
		break ;
#endif //_USE_PCE
	}

	emc = &emu_config[ emu_mode  ] ;

#ifdef _OVER_CLOCK
	switch( emc->cpu_freq )
	{
	default :
	case CPU_FREQ_222: scePowerSetClockFrequency(222,222,111); break ;
	case CPU_FREQ_266: scePowerSetClockFrequency(266,266,133); break ;
	case CPU_FREQ_333: scePowerSetClockFrequency(333,333,166); break ;
	}
#endif

_UPDATE_CFG :
	switch( emc->sound_vol )
	{
	case VOL_08 : psp_sound_vol = 0x8000 ; break ;
	case VOL_10 : psp_sound_vol = 0xA000 ; break ;
	case VOL_12 : psp_sound_vol = 0xC000 ; break ;
	case VOL_14 : psp_sound_vol = 0xE000 ; break ;
	case VOL_16 : psp_sound_vol = 0xFFFF ; break ;
	}

	fps      = 0 ;
	fps_last = 0 ;
	skip     = 0 ;
	skip_s   = emc->skip_count ;
	tm_ed    = sceKernelLibcClock() ;
	tm_run   = TM_1FRAME ;
	tm_skip  = TM_1FRAME ;
	tm_res   = 0 ;

	key_menu = menu_keys[ EmuSysCfg_Get()->key_menu ] ;

	sceKernelDcacheWritebackAll();

	switch( emu_mode )
	{
	default : goto _MAIN_MENU ;

#ifdef _USE_MD
	case EMU_MD  : 
		if( !md_is_loadrom() ){ goto _MAIN_MENU ; }
		md_setconfig( emc ) ;
		PSP_Sound_Enable(0,emc->sound_on) ;
		switch( emc->screen_mode )
		{
		case SCREEN_FIT  : goto _MD_LOOP_FIT  ;
		case SCREEN_FULL : goto _MD_LOOP_FULL ;
		}
		goto _MD_LOOP  ;
#endif //_USE_MD

#ifdef _USE_PCE
	case EMU_PCE : 
		pce_setconfig( emc ) ;
		PSP_Sound_Enable(0,emc->sound_on) ;
		goto _PCE_LOOP ;
#endif //_USE_PCE
	}

#ifdef _USE_MD
	for(;;)
	{
_MD_LOOP :

		switch( UpdateTimer() )
		{
		case 1 : goto _UPDATE_CFG ;
		case 2 : goto _MAIN_MENU  ;
		}

		if( (skip == 0) || (skip_s == 0) )
		{
			md_setframe( PSP_Screen_GetBackBuffer( 80, 16 ), PSP_LINESIZE * 2 ) ;
			md_run() ;

			if( exec_cfg > 0 ){ ShowExecCfgMsg(); }
			else if( emc->show_fps ){ ShowFPS( emc->show_fps, fps_last ) ;  }
			if( emc->show_flags & SHOW_VSYNC ){ PSP_Screen_WaitVsyncFast() ; }
			PSP_Screen_Flip() ;
			PSP_Power_CheckSleep() ;
		
			tm_ed  = sceKernelLibcClock() ;
			tm_run = tm_ed - tm_st ;
			skip   = calc_skip( tm_run ) ;
		}
		else
		{
			md_run_skip() ;

			tm_ed   = sceKernelLibcClock() ;
			tm_skip = tm_ed - tm_st ;
			skip    = calc_skip( tm_skip ) ;
		}
	}

	for(;;)
	{
_MD_LOOP_FIT :

		switch( UpdateTimer() )
		{
		case 1 : goto _UPDATE_CFG ;
		case 2 : goto _MAIN_MENU  ;
		}

		if( (skip == 0) || (skip_s == 0) )
		{
			if( exec_cfg > 0 ){ ShowExecCfgMsg2(); }
			else if( emc->show_fps ){ ShowFps2( emc->show_fps, fps_last ) ;  }
			if( emc->show_flags & SHOW_VSYNC ){ PSP_Screen_WaitVsyncFast() ; }
			PSP_Screen_Flip() ;
			PSP_Screen_TempFlip() ;
			md_setframe( PSP_Screen_GetTempBuffer( 8, 8 ), PSP_LINESIZE * 2 ) ;
			md_run() ;

			PSP_Screen_BitBltGe( 
				PSP_Screen_GetTempBuffer( 8, 16 ), 
				320, 224,
				(480 - 388)/2, 0, (480 - 388)/2 + 388, PSP_SCREEN_HEIGHT ) ;

			//PSP_Screen_Flip() ;
			PSP_Power_CheckSleep() ;
		
			tm_ed  = sceKernelLibcClock() ;
			tm_run = tm_ed - tm_st ;
			skip   = calc_skip( tm_run ) ;
		}
		else
		{
			md_run_skip() ;

			tm_ed   = sceKernelLibcClock() ;
			tm_skip = tm_ed - tm_st ;
			skip    = calc_skip( tm_skip ) ;
		}
	}

	for(;;)
	{
_MD_LOOP_FULL :

		switch( UpdateTimer() )
		{
		case 1 : goto _UPDATE_CFG ;
		case 2 : goto _MAIN_MENU  ;
		}

		if( (skip == 0) || (skip_s == 0) )
		{
			if( exec_cfg > 0 ){ ShowExecCfgMsg2(); }
			else if( emc->show_fps ){ ShowFps2( emc->show_fps, fps_last ) ;  }
			if( emc->show_flags & SHOW_VSYNC ){ PSP_Screen_WaitVsyncFast() ; }
			PSP_Screen_Flip() ;
			PSP_Screen_TempFlip() ;
			md_setframe( PSP_Screen_GetTempBuffer( 16, 8 ), PSP_LINESIZE * 2 ) ;
			md_run() ;

			if( md_is_width320() )
			{
				PSP_Screen_BitBltGe( 
					PSP_Screen_GetTempBuffer( 16, 16 ), 
					320, 224,
					0, 0, PSP_SCREEN_WIDTH, PSP_SCREEN_HEIGHT ) ;
			}
			else
			{
				PSP_Screen_BitBltGe( 
					PSP_Screen_GetTempBuffer( 48, 16 ), 
					256, 224,
					0, 0, PSP_SCREEN_WIDTH, PSP_SCREEN_HEIGHT ) ;
			}

			PSP_Power_CheckSleep() ;
		
			tm_ed  = sceKernelLibcClock() ;
			tm_run = tm_ed - tm_st ;
			skip   = calc_skip( tm_run ) ;
		}
		else
		{
			md_run_skip() ;

			tm_ed   = sceKernelLibcClock() ;
			tm_skip = tm_ed - tm_st ;
			skip    = calc_skip( tm_skip ) ;
		}
	}

#endif //_USE_MD

#ifdef _USE_PCE
	for(;;)
	{
_PCE_LOOP :
		fps++ ;
		tm_ed = sceKernelLibcClock() ;
		if( (tm_ed - tm_st) > (TM_1FRAME*60) )
		{
			if( emc->show_fps ){ ShowFPS( emc->show_fps, fps ) ;  }
			fps   = 0 ;
			fps_s = 0 ;
			tm_st = tm_ed ;
		}

		PSP_Ctrlpad_Update() ;

		pce_setpaddata( PSP_Ctrlpad_GetButtons() ) ;
		if( (skip == 0) || (emc->skip_count == 0) )
		{
			pce_setframe( PSP_Screen_GetBackBuffer( 80-8, 24 ), PSP_LINESIZE * 2 ) ;
			pce_run() ;
			if( emc->sound_on )
			{
//				pce_sound_update( PSP_Sound_GetBackBuffer(0), PSP_SOUND_SAMPLE ) ;
//				PSP_Sound_BufferFlip(0) ;
			}

			if( emc->show_flags & SHOW_VSYNC ){ PSP_Screen_WaitVsync() ; }
			PSP_Screen_Flip() ;
			PSP_Power_CheckSleep() ;
			skip++;
			fps_s++;
		}
		else
		{
			pce_run_skip() ;
			if( emc->sound_on )
			{
				pce_sound_update( PSP_Sound_GetBackBuffer(0), MAX_PSP_SOUND_BUFFER/2 ) ;
				PSP_Sound_BufferFlip(0) ;
			}
			skip++;
			if( skip >= emc->skip_count ){ skip = 0 ; }
		}

		if( PSP_Ctrlpad_GetButtons() & CTRL_LTRIGGER ){ goto _MAIN_MENU ; }

	}
#endif //_USE_PCE

	return 0 ;
}

///////////////////////////////////////////////////////////////////////////
/// PSP Sound
///////////////////////////////////////////////////////////////////////////
#define MAX_PSP_SOUND_CHANNEL  (1)
#define MAX_PSP_VOLUME         (0x8000)
#define MAX_PSP_SOUND_BANK     (6)
#define PSP_SOUND_SAMPLE_W     (PSP_SOUND_SAMPLE/2)

volatile int   psp_sound_terminate = 0 ;
unsigned long  psp_sound_handle [MAX_PSP_SOUND_CHANNEL] ;
unsigned long  psp_sound_thid   [MAX_PSP_SOUND_CHANNEL] ;
unsigned short psp_sound_enable [MAX_PSP_SOUND_CHANNEL] ;
static short   psp_sound_playbuf[MAX_PSP_SOUND_CHANNEL][2][ PSP_SOUND_SAMPLE * 2 ];
/*
static short   psp_sound_buffer [MAX_PSP_SOUND_CHANNEL][MAX_PSP_SOUND_BANK][MAX_PSP_SOUND_BUFFER] ;
short          psp_sound_ptr    [MAX_PSP_SOUND_CHANNEL] ;
unsigned long  psp_sound_bank_p [MAX_PSP_SOUND_CHANNEL] ;
unsigned long  psp_sound_bank_w [MAX_PSP_SOUND_CHANNEL] ;
unsigned char  psp_sound_bank_f [MAX_PSP_SOUND_CHANNEL][MAX_PSP_SOUND_BANK] ;
unsigned long  psp_sound_bankuse[MAX_PSP_SOUND_CHANNEL][MAX_PSP_SOUND_BANK] ;
unsigned long  psp_sound_handle_w ;
//*/

static int PSP_Sound_Thread( int args, void *argp )
{
	int nChannel = *((int*)(argp)) ;
	short outbank = 0 ;
#ifdef _DEBUG
	int n ;
	unsigned long tm_sound ;
#endif //_DEBUG

	while ( psp_sound_terminate == 0 ) 
	{
#ifdef _DEBUG
		tm_sound = sceKernelLibcClock() ;
#endif //_DEBUG

		if( psp_sound_enable[ nChannel ] )
		{
			outbank ^= 1 ;

#ifdef _DEBUG

			//memset( psp_sound_playbuf[nChannel][outbank], 0x55, PSP_SOUND_SAMPLE * sizeof(short) * 16 ) ;

			md_sound_update( psp_sound_playbuf[nChannel][outbank], PSP_SOUND_SAMPLE ) ;

			tm_debug[2] = sceKernelLibcClock() - tm_sound ;

			/*
			tm_debug[6] = 0 ;
			for( n=PSP_SOUND_SAMPLE*2; n < PSP_SOUND_SAMPLE * 16; n++ )
			{
				if( psp_sound_playbuf[nChannel][outbank][n] != 0x5555 )
				{
					tm_debug[6]++ ;
				}
				//else{ break ; }
			}
			*/

#else  //
			md_sound_update( psp_sound_playbuf[nChannel][outbank], PSP_SOUND_SAMPLE ) ;

#endif //_DEBUG
			//sceAudioOutputPannedBlocking( psp_sound_handle[nChannel], MAX_PSP_VOLUME, MAX_PSP_VOLUME, psp_sound_playbuf[nChannel][outbank] );
			sceAudioOutputPannedBlocking( psp_sound_handle[nChannel], psp_sound_vol, psp_sound_vol, psp_sound_playbuf[nChannel][outbank] );
		}
		else
		{
			outbank ^= 1 ;
			memset4x( (unsigned long*)psp_sound_playbuf[nChannel][outbank], 0, PSP_SOUND_SAMPLE ) ;
#ifdef _DEBUG
			tm_debug[2] = sceKernelLibcClock() - tm_sound ;
#endif //_DEBUG
			sceAudioOutputPannedBlocking( psp_sound_handle[nChannel], 0, 0, psp_sound_playbuf[nChannel][outbank] );
		}
	}

	sceKernelExitThread(0);
	return 0 ;
}

// nChannel : 0 - 2
int PSP_Sound_Init( int nChannel )
{
	int n ;

	char thread_name[] = "pgasndX" ;

	memset( psp_sound_playbuf, 0, sizeof(psp_sound_playbuf) ) ;

	/*
	for( n=0; n < MAX_PSP_SOUND_BANK ; n++ )
	{
		memset( psp_sound_buffer[0][n], 0, MAX_PSP_SOUND_BUFFER * sizeof(short) ) ;
	}
	*/

	psp_sound_terminate = 0 ;

	if( nChannel > MAX_PSP_SOUND_CHANNEL ){ nChannel = MAX_PSP_SOUND_CHANNEL ; }

	for( n=0; n < MAX_PSP_SOUND_CHANNEL ; n++ )
	{
		psp_sound_handle[n] = -1 ;
		psp_sound_thid  [n] = -1 ;
		psp_sound_enable[n] = 0  ;
	}

//	psp_sound_handle_w = sceAudio_3( -1, PSP_SOUND_SAMPLE_W, 0 ) ;

	for( n=0; n < nChannel ; n++ )
	{
		psp_sound_handle[n] = sceAudioChReserve( -1, PSP_SOUND_SAMPLE, 0 ) ;

		if( psp_sound_handle[n] >= 0 )
		{
			thread_name[6] = '0' + n ;
			psp_sound_thid[n] = sceKernelCreateThread(
									thread_name,
									&PSP_Sound_Thread,
									0x12,0x10000,0,NULL );

			if( psp_sound_thid[n] >= 0 )
			{
				sceKernelStartThread( psp_sound_thid[n], sizeof(n), &n );
			}
		}
	}

	return n ;
}

void PSP_Sound_Enable( short nCh, short nEnable )
{
	psp_sound_enable[ nCh ] = nEnable ;
}


void PSP_Sound_Term(void)
{
	int n ; 

	psp_sound_terminate = 1 ;

	for( n=0; n < MAX_PSP_SOUND_CHANNEL ; n++ )
	{
		if( psp_sound_thid[n] >= 0 )
		{
			sceKernelWaitThreadEnd(psp_sound_thid[n],NULL);
			sceKernelDeleteThread (psp_sound_thid[n]);

			psp_sound_thid[n] = -1 ;
		}

		if( psp_sound_handle[n] >= 0 )
		{
			sceAudioChRelease( psp_sound_handle[n] );
		}
	}

}

///////////////////////////////////////////////////////////////////////////
/// PSP Ctrlpad
///////////////////////////////////////////////////////////////////////////
PSP_PAD_DATA  psp_pad_data ;

int PSP_Ctrlpad_Init( void )
{
	sceCtrlSetSamplingCycle(0);
	sceCtrlSetSamplingMode (1);
}

void PSP_Ctrlpad_Update( void )
{
	sceCtrlReadBufferPositive( &psp_pad_data, 1 );
}

unsigned long PSP_Ctrlpad_GetButtons( void )
{
	return psp_pad_data.buttons ;
}

PSP_PAD_DATA* PSP_Ctrlpad_GetPadData()
{
	return &psp_pad_data ;
}

///////////////////////////////////////////////////////////////////////////
/// PSP Power
///////////////////////////////////////////////////////////////////////////
int bSleep=0;

int PSP_Power_CheckSleep(void)
{
	int n ;

	if( bSleep )
	{
		PSP_Sound_Enable(0,0) ;

		while( bSleep )
		{
			sceDisplayWaitVblankStart();
		}

		PSP_Sound_Enable(0,emc->sound_on) ;

		return 1 ;
	}

	return 0 ;
}

////////////////////////////////////
/// from RIN.  thanks for MIRAKICHI.
////////////////////////////////////
#define POWER_CB_POWER		0x80000000
#define POWER_CB_HOLDON		0x40000000
#define POWER_CB_STANDBY	0x00080000
#define POWER_CB_RESCOMP	0x00040000
#define POWER_CB_RESUME		0x00020000
#define POWER_CB_SUSPEND	0x00010000
#define POWER_CB_EXT		0x00001000
#define POWER_CB_BATLOW		0x00000100
#define POWER_CB_BATTERY	0x00000080
#define POWER_CB_BATTPOWER	0x0000007F

inline void on_exit_game()
{
#ifdef _OVER_CLOCK
		scePowerSetClockFrequency(222,222,111);
#endif

	sceKernelPowerLock(0);
	EmuSysCfg_Save( path_exec ) ;
	if( strlen( path_cfg   ) > 0                  ){ EmuCfg_Save      ( emc       , path_cfg   ) ; }
	if( strlen( path_cheat ) > 0                  ){ EmuCheat_Save    ( &emu_cheat, path_cheat ) ; }
	if( strlen( path_sram  ) > 0 && emc->autosram ){ EmuCore_SaveSRAM ( path_sram              ) ; }
	if( strlen( path_state ) > 0                  ){ EmuState_Save    ( &emu_state, path_state ) ; }
	sceKernelPowerUnlock(0);
}
/*
// z[{^IɃR[obN
int exit_callback(void)
{
	bSleep=1;

	on_exit_game() ;

	sceKernelExitGame();
	return 0;
}

// X[vsɃR[obN
void power_callback(int unknown, int pwrflags)
{
	if(pwrflags & POWER_CB_POWER)
	{
		if( !bSleep )
		{
			bSleep=1;

			on_exit_game() ;
		}
	}
	else if(pwrflags & POWER_CB_BATLOW)
	{
		//if( EmuSysCfg_Get()->option_flags & OPTION_LOWBATT )
		{
			if( !bSleep )
			{
				bSleep=1;

				on_exit_game() ;

				scePowerRequestSuspend() ;
			}
		}
	}
	else if(pwrflags & POWER_CB_RESCOMP)
	{
		bSleep=0;

#ifdef _OVER_CLOCK
		switch( emc->cpu_freq )
		{
		default :
		case CPU_FREQ_222: scePowerSetClockFrequency(222,222,111); break ;
		case CPU_FREQ_266: scePowerSetClockFrequency(266,266,133); break ;
		case CPU_FREQ_333: scePowerSetClockFrequency(333,333,166); break ;
		}
#endif

	}

	// R[obN֐̍ēo^
	// ixĂ΂ꂽēo^ƂȂƎɃR[obNȂj
	int cbid = sceKernelCreateCallback("Power Callback", power_callback);
	scePowerRegisterCallback(0, cbid);
}

// |[OpXbh
int CallbackThread(int args, void *argp)
{
	int cbid;
	
	// R[obN֐̓o^
	cbid = sceKernelCreateCallback("Exit Callback", exit_callback);
	sceKernelRegisterExitCallback(cbid);
	cbid = sceKernelCreateCallback("Power Callback", power_callback);
	scePowerRegisterCallback(0, cbid);
	
	// |[O
	sceKernelPollCallbacks();
}

int PSP_Power_Init(void)
{
	int thid = 0;
	
	// |[OpXbh̐
	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
		sceKernelStartThread(thid, 0, 0);
	
	return thid;
}
*/
///////////////////////////////////////////////////////////////////////////

